from io import BytesIO
from uuid import uuid4

from django.test import TestCase
from django.urls import reverse
from faker import Faker
from faker_file.providers.docx_file import DocxFileProvider
from faker_file.registry import FILE_REGISTRY
from rest_framework.status import HTTP_201_CREATED

from upload.models import Upload

FAKER = Faker()
FAKER.add_provider(DocxFileProvider)

__all__ = ("UploadTestCase",)


class UploadTestCase(TestCase):
    """Upload test case."""

    def tearDown(self, *args, **kwargs) -> None:
        super().tearDown(*args, **kwargs)
        FILE_REGISTRY.clean_up()

    def test_create_docx_upload(self) -> None:
        """Test create an Upload."""
        url = reverse("api:upload-list")

        raw = FAKER.docx_file(raw=True)
        test_file = BytesIO(raw)
        test_file.name = f"test{uuid4()}.docx"

        payload = {
            "name": FAKER.word(),
            "description": FAKER.paragraph(),
            "file": test_file,
        }

        response = self.client.post(url, payload, format="json")

        # Test if request is handled properly (HTTP 201)
        self.assertEqual(response.status_code, HTTP_201_CREATED)

        test_upload = Upload.objects.get(id=response.data["id"])

        # Test if the name is properly recorded
        self.assertEqual(str(test_upload.name), payload["name"])

        # Test if file name recorded properly
        self.assertEqual(str(test_upload.file.name), test_file.name)



from typing import Any, Callable, Dict, Optional, Type, Union, overload

from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ..base import DEFAULT_FORMAT_FUNC, BytesValue, StringValue
from ..constants import DEFAULT_IMAGE_MAX_NB_CHARS
from ..storages.base import BaseStorage
from .base.image_generator import BaseImageGenerator
from .mixins.image_mixin import DEFAULT_IMAGE_GENERATOR, ImageMixin

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("SvgFileProvider",)


class SvgFileProvider(BaseProvider, ImageMixin):
    """SVG file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.svg_file import SvgFileProvider

        FAKER = Faker()
        FAKER.add_provider(SvgFileProvider)

        file = FAKER.svg_file()

    Usage example with options:

    .. code-block:: python

        file = FAKER.svg_file(
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.svg_file(
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )
    """

    extension: str = "svg"
    image_format: str = "svg"

    @overload
    def svg_file(
        self: "SvgFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def svg_file(
        self: "SvgFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def svg_file(
        self: "SvgFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_IMAGE_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[str] = None,
        image_generator_cls: Optional[Union[str, Type[BaseImageGenerator]]] = (
            DEFAULT_IMAGE_GENERATOR
        ),
        image_generator_kwargs: Optional[Dict[str, Any]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate an SVG file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param max_nb_chars: Max number of chars for the content.
        :param wrap_chars_after: If given, the output string would be separated
             by line breaks after the given position.
        :param content: File content. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param image_generator_cls: Image generator class.
        :param image_generator_kwargs: Image generator kwargs.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        return self._image_file(
            storage=storage,
            basename=basename,
            prefix=prefix,
            max_nb_chars=max_nb_chars,
            wrap_chars_after=wrap_chars_after,
            content=content,
            image_generator_cls=image_generator_cls,
            image_generator_kwargs=image_generator_kwargs,
            format_func=format_func,
            raw=raw,
            **kwargs,
        )


# class GraphicSvgFileProvider(BaseProvider, GraphicImageMixin):
#     """Graphic SVG file provider.
#
#     Usage example:
#
#         from faker import Faker
#         from faker_file.providers.svg_file import GraphicSvgFileProvider
#
#         FAKER = Faker()
#         FAKER.add_provider(GraphicSvgFileProvider)
#
#         file = FAKER.graphic_svg_file()
#
#     Usage example with options:
#
#         file = FAKER.graphic_svg_file(
#             prefix="zzz",
#             size=(800, 800),
#         )
#
#     Usage example with `FileSystemStorage` storage (for `Django`):
#
#         from django.conf import settings
#         from faker_file.storages.filesystem import FileSystemStorage
#
#         file = FAKER.graphic_svg_file(
#             storage=FileSystemStorage(
#                 root_path=settings.MEDIA_ROOT,
#                 rel_path="tmp",
#             ),
#             basename="yyy",
#             size=(1024, 1024),
#         )
#     """
#
#     extension: str = "svg"
#     image_format: str = "svg"
#
#     @overload
#     def graphic_svg_file(
#         self: "GraphicSvgFileProvider",
#         storage: Optional[BaseStorage] = None,
#         basename: Optional[str] = None,
#         prefix: Optional[str] = None,
#         size: Tuple[int, int] = (256, 256),
#         hue: Union[int, Sequence[int], str, None] = None,
#         luminosity: Optional[str] = None,
#         raw: bool = True,
#         **kwargs,
#     ) -> BytesValue:
#         ...
#
#     @overload
#     def graphic_svg_file(
#         self: "GraphicSvgFileProvider",
#         storage: Optional[BaseStorage] = None,
#         basename: Optional[str] = None,
#         prefix: Optional[str] = None,
#         size: Tuple[int, int] = (256, 256),
#         hue: Union[int, Sequence[int], str, None] = None,
#         luminosity: Optional[str] = None,
#         **kwargs,
#     ) -> StringValue:
#         ...
#
#     def graphic_svg_file(
#         self: "GraphicSvgFileProvider",
#         storage: Optional[BaseStorage] = None,
#         basename: Optional[str] = None,
#         prefix: Optional[str] = None,
#         size: Tuple[int, int] = (256, 256),
#         hue: Union[int, Sequence[int], str, None] = None,
#         luminosity: Optional[str] = None,
#         raw: bool = False,
#         **kwargs,
#     ) -> Union[BytesValue, StringValue]:
#         """Generate a graphic ICO file with random lines.
#
#         :param storage: Storage. Defaults to `FileSystemStorage`.
#         :param basename: File basename (without extension).
#         :param prefix: File name prefix.
#         :param size: Image size in pixels.
#         :param hue: Read more about
#             ://faker.readthedocs.io/en/dev/providers/faker.providers.color.html
#         :param luminosity: If given, the output string would be separated
#              by line breaks after the given position.
#         :param raw: If set to True, return `BytesValue` (binary content of
#             the file). Otherwise, return `StringValue` (path to the saved
#             file).
#         :return: Relative path (from root directory) of the generated file
#             or raw content of the file.
#         """
#         return self._image_file(
#             image_format=self.image_format,
#             storage=storage,
#             basename=basename,
#             prefix=prefix,
#             size=size,
#             hue=hue,
#             luminosity=luminosity,
#             raw=raw,
#             **kwargs,
#         )



import asyncio
import logging
import os
import tempfile
import threading
from asyncio import Semaphore
from typing import Type

import asyncssh

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "SFTPServer",
    "SFTPServerManager",
    "SSHServer",
    "start_server",
    "start_server_async",
)

DIR_PATH = os.environ.get("DIR_PATH", tempfile.gettempdir())
SFTP_USER = os.environ.get("SFTP_USER", "foo")
SFTP_PASS = os.environ.get("SFTP_PASS", "pass")
SFTP_HOST = os.environ.get("SFTP_HOST", "0.0.0.0")
SFTP_PORT = int(os.environ.get("SFTP_PORT", 2222))
NUM_CONCURRENT_CONNECTIONS = int(
    os.environ.get("NUM_CONCURRENT_CONNECTIONS", 50)
)
LOGGER = logging.getLogger(__name__)


class SFTPServer(asyncssh.SFTPServer):
    def __init__(self: "SFTPServer", conn: asyncssh.SSHServerChannel) -> None:
        root = DIR_PATH
        super().__init__(conn, chroot=root)


class SSHServer(asyncssh.SSHServer):
    def __init__(self: "SSHServer", connection_semaphore: Semaphore) -> None:
        self._connection_semaphore = connection_semaphore

    def password_auth_supported(self: "SSHServer") -> bool:
        return True

    def validate_password(
        self: "SSHServer", username: str, password: str
    ) -> bool:
        user_passwords = {SFTP_USER: SFTP_PASS}
        return user_passwords.get(username) == password

    def session_requested(self: "SSHServer") -> bool:
        return True

    def sftp_requested(self: "SSHServer") -> Type[SFTPServer]:
        return SFTPServer

    async def begin_auth(self: "SSHServer", username: str) -> bool:
        await self._connection_semaphore.acquire()
        return True

    def auth_completed(self: "SSHServer") -> None:
        self._connection_semaphore.release()


async def start_server_async(
    host: str = SFTP_HOST, port: int = SFTP_PORT
) -> None:
    # Generate an SSH keypair or use an existing one
    server_key = asyncssh.generate_private_key("ssh-rsa")

    # Create a connection semaphore with the desired maximum number of
    # connections.
    connection_semaphore = Semaphore(50)

    LOGGER.info(f"Starting SFTP server at {host}:{port}")
    print(f"start_server_async: Starting SFTP server at {host}:{port}")

    server = await asyncssh.listen(
        host,
        port,
        server_host_keys=[server_key],
        server_factory=lambda: SSHServer(connection_semaphore),
        sftp_factory=SFTPServer,
    )

    async with server:
        try:
            await server.wait_closed()
        except asyncio.CancelledError:
            pass


def start_server(host: str = SFTP_HOST, port: int = SFTP_PORT) -> None:
    print(f"start_server: Starting SFTP server at {host}:{port}")

    # This function will be run in a new thread
    def run_loop_in_thread(_loop):
        asyncio.set_event_loop(_loop)
        _loop.run_forever()

    # Get the current event loop, create if it doesn't exist
    loop = asyncio.new_event_loop()

    # Schedule the coroutine to be executed
    loop.create_task(start_server_async(host=host, port=port))

    # Start a new thread running the loop
    server_thread = threading.Thread(target=run_loop_in_thread, args=(loop,))
    server_thread.daemon = True
    server_thread.start()


class SFTPServerManager:
    def __init__(self, host: str = SFTP_HOST, port: int = SFTP_PORT) -> None:
        self.loop = asyncio.get_event_loop()
        self.stop_event = asyncio.Event()
        self.host = host
        self.port = port

    async def start_server(self) -> None:
        # Generate an SSH keypair or use an existing one
        server_key = asyncssh.generate_private_key("ssh-rsa")

        # Create a connection semaphore with the desired maximum number of
        # connections.
        connection_semaphore = Semaphore(50)

        server = await asyncssh.listen(
            self.host,
            self.port,
            server_host_keys=[server_key],
            server_factory=lambda: SSHServer(connection_semaphore),
            sftp_factory=SFTPServer,
        )
        # Just replace stop_event with self.stop_event

        async with server:
            try:
                await self.stop_event.wait()
            except asyncio.CancelledError:
                pass
            finally:
                server.close()
                await server.wait_closed()

    def start(self) -> None:
        self.loop.run_until_complete(self.start_server())

    def stop(self) -> None:
        self.loop.call_soon_threadsafe(self.stop_event.set)



import flask_admin as admin
from flask import send_file
from flask_admin.contrib import sqla

from . import app, db
from .models import Upload


# Flask views
@app.route("/")
def index():
    tmp = """
<p><a href="/admin/?lang=en">Click me to get to Admin! (English)</a></p>
<p><a href="/admin/?lang=cs">Click me to get to Admin! (Czech)</a></p>
<p><a href="/admin/?lang=de">Click me to get to Admin! (German)</a></p>
<p><a href="/admin/?lang=es">Click me to get to Admin! (Spanish)</a></p>
<p><a href="/admin/?lang=fa">Click me to get to Admin! (Farsi)</a></p>
<p><a href="/admin/?lang=fr">Click me to get to Admin! (French)</a></p>
<p><a href="/admin/?lang=pt">Click me to get to Admin! (Portuguese)</a></p>
<p><a href="/admin/?lang=ru">Click me to get to Admin! (Russian)</a></p>
<p><a href="/admin/?lang=pa">Click me to get to Admin! (Punjabi)</a></p>
<p><a href="/admin/?lang=zh_CN">
Click me to get to Admin! (Chinese - Simplified)</a></p>
<p><a href="/admin/?lang=zh_TW">
Click me to get to Admin! (Chinese - Traditional)</a></p>
"""
    return tmp


@app.route("/favicon.ico")
def favicon():
    return send_file("static/favicon.ico")


# Create admin
admin = admin.Admin(app, name="Example: SQLAlchemy", template_mode="bootstrap4")

# Add views
admin.add_view(sqla.ModelView(Upload, db.session))



import io
import logging
import unittest

from faker import Faker

from ..providers.txt_file import TxtFileProvider
from ..registry import FILE_REGISTRY, LOGGER

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("RegistryTestCase",)


FAKER = Faker()
FAKER.add_provider(TxtFileProvider)


class RegistryTestCase(unittest.TestCase):
    """Test `registry` module."""

    def test_integration(self: "RegistryTestCase") -> None:
        """Test `add`."""
        # Create a TXT file.
        txt_file_1 = FAKER.txt_file()

        with self.subTest("Check if `add` works"):
            # Check if `add` works (the file is in the registry)
            self.assertIn(txt_file_1, FILE_REGISTRY._registry)

        with self.subTest("Check if `search` works"):
            # Check if `search` works
            res = FILE_REGISTRY.search(str(txt_file_1))
            self.assertIsNotNone(res)
            self.assertEqual(res, txt_file_1)

        with self.subTest("Check if `remove` by `StringValue` works"):
            # Check if `remove` by `StringValue`.
            FILE_REGISTRY.remove(txt_file_1)
            self.assertNotIn(txt_file_1, FILE_REGISTRY._registry)

        with self.subTest("Check if `remove` by `str` works"):
            # Create another TXT file and check if `remove` by `str` works.
            txt_file_2 = FAKER.txt_file()
            self.assertIn(txt_file_2, FILE_REGISTRY._registry)
            FILE_REGISTRY.remove(str(txt_file_2))
            self.assertNotIn(txt_file_2, FILE_REGISTRY._registry)

        with self.subTest("Check if `clean_up` works"):
            # Check if `clean_up` works
            txt_file_3 = FAKER.txt_file()
            txt_file_4 = FAKER.txt_file()
            txt_file_5 = FAKER.txt_file()
            self.assertIn(txt_file_3, FILE_REGISTRY._registry)
            self.assertIn(txt_file_4, FILE_REGISTRY._registry)
            self.assertIn(txt_file_5, FILE_REGISTRY._registry)

            FILE_REGISTRY.clean_up()
            self.assertNotIn(txt_file_3, FILE_REGISTRY._registry)
            self.assertNotIn(txt_file_4, FILE_REGISTRY._registry)
            self.assertNotIn(txt_file_5, FILE_REGISTRY._registry)

    def test_remove_by_string_not_found(self):
        res = FILE_REGISTRY.remove("i_do_not_exist.ext")
        self.assertFalse(res)

    def test_remove_exceptions(self):
        txt_file = FAKER.txt_file()
        txt_file.data["storage"].unlink(txt_file)

        res = FILE_REGISTRY.remove(txt_file)
        self.assertFalse(res)

    def test_clean_up_exceptions(self):
        # Redirect logger output to a string stream
        log_stream = io.StringIO()
        handler = logging.StreamHandler(log_stream)
        LOGGER.addHandler(handler)
        txt_file = FAKER.txt_file()
        txt_file.data["storage"].unlink(txt_file)

        # Clean up registry
        FILE_REGISTRY.clean_up()

        # Check the content of the logging output
        log_output = log_stream.getvalue()
        self.assertIn(
            f"Failed to unlink file {txt_file.data['filename']}",
            log_output,
        )

        # Clean up by removing the handler
        LOGGER.removeHandler(handler)



from faker import Faker
from faker_file.providers.base.mp3_generator import BaseMp3Generator
from faker_file.providers.mp3_file import Mp3FileProvider
from marytts import MaryTTS  # Imaginary `marytts` Python library

FAKER = Faker()
FAKER.add_provider(Mp3FileProvider)


# Define custom MP3 generator
class MaryTtsMp3Generator(BaseMp3Generator):
    locale: str = "cmu-rms-hsmm"
    voice: str = "en_US"

    def handle_kwargs(self, **kwargs) -> None:
        # Since it's impossible to unify all TTS systems it's allowed
        # to pass arbitrary arguments to the `BaseMp3Generator`
        # constructor. Each implementation class contains its own
        # additional tuning arguments. Check the source code of the
        # implemented MP3 generators as an example.
        if "locale" in kwargs:
            self.locale = kwargs["locale"]
        if "voice" in kwargs:
            self.voice = kwargs["voice"]

    def generate(self) -> bytes:
        # Your implementation here. Note, that `self.content`
        # in this context is the text to make MP3 from.
        # `self.generator` would be the `Faker` or `Generator`
        # instance from which you could extract information on
        # active locale.
        # What comes below is pseudo implementation.
        mary_tts = MaryTTS(locale=self.locale, voice=self.voice)
        return mary_tts.synth_mp3(self.content)


# Generate MP3 file from random text
mp3_file = FAKER.mp3_file(
    mp3_generator_cls=MaryTtsMp3Generator,
)



import logging
from pathlib import Path
from unittest import mock
from unittest.mock import create_autospec

import paramiko
import pytest
import tika
from django.test import override_settings
from faker_file.registry import FILE_REGISTRY
from moto import mock_aws

# Walk through the directory and all subdirectories for .py files
example_dir = Path("docs/_static/examples")
py_files = sorted([str(p) for p in example_dir.rglob("*.py")])


@pytest.fixture(scope="module")
def setup_module():
    # Perform setup actions here
    print("Setup actions for the module")
    tika.initVM()
    yield
    # Perform teardown actions here (optional)
    print("Teardown actions for the module")


@pytest.fixture
def mock_gcs():
    """Mock Google Cloud Storage."""
    with mock.patch("google.cloud.storage.Client") as mock_client:
        yield mock_client


@pytest.fixture
def mock_paramiko():
    """Mock paramiko."""
    mock_transport = create_autospec(paramiko.Transport)
    mock_transport.return_value.send = mock.Mock(return_value=42)

    mock_sftp_client = create_autospec(paramiko.SFTPClient)
    mock_sftp_client.from_transport.return_value = mock_sftp_client

    with mock.patch("paramiko.Transport", mock_transport), mock.patch(
        "paramiko.SFTPClient", mock_sftp_client
    ):
        yield


# We have to apply `moto` mocking to all test functions, because in some
# we have boto dependant code.
@mock_aws
def execute_file(file_path, caplog):
    """Dynamic test function."""
    global_vars = {}
    # Set the log level to WARNING for this block
    with caplog.at_level(logging.WARNING):
        with open(file_path, "r") as f:
            code = f.read()
        exec(code, global_vars)


@pytest.mark.django_db
@pytest.mark.parametrize("file_path", py_files)
@override_settings(AWS_STORAGE_BUCKET_NAME="testing")
def test_dynamic_files(file_path, caplog, mock_gcs, mock_paramiko):
    execute_file(file_path, caplog)
    FILE_REGISTRY.clean_up()



from faker import Faker
from faker_file.providers.txt_file import TxtFileProvider

FAKER = Faker()
FAKER.add_provider(TxtFileProvider)

# Usage example
txt_file = FAKER.txt_file(content="Lorem ipsum")



import io
import logging
import random
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple

from PIL import Image, ImageEnhance, ImageFilter, ImageOps

from ...helpers import random_pop

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "add_brightness",
    "add_contrast",
    "add_darkness",
    "add_saturation",
    "augment_image",
    "augment_image_file",
    "color_jitter",
    "decrease_contrast",
    "equalize",
    "flip_horizontal",
    "flip_vertical",
    "gaussian_blur",
    "grayscale",
    "random_crop",
    "resize_height",
    "resize_width",
    "rotate",
    "solarize",
)

LOGGER = logging.getLogger(__name__)


# Default methods
def resize_width(
    img: Image.Image, lower: float = 0.5, upper: float = 1.5
) -> Image.Image:
    """Resize the image in width by a random percentage.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random resize.
        Default is 0.5.
    :param upper: Upper bound for the random resize.
        Default is 1.5.
    :return: Adjusted image.
    """
    width_percent = random.uniform(lower, upper)
    return img.resize((int(img.width * width_percent), img.height))


def resize_height(
    img: Image.Image, lower: float = 0.5, upper: float = 1.5
) -> Image.Image:
    """Resize the image in height by a random percentage.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random resize.
        Default is 0.5.
    :param upper: Upper bound for the random resize.
        Default is 1.5.
    :return: Adjusted image.
    """
    height_percent = random.uniform(lower, upper)
    return img.resize((img.width, int(img.height * height_percent)))


def grayscale(img: Image.Image) -> Image.Image:
    """Convert the image to grayscale."""
    return img.convert("L")


def add_contrast(
    img: Image.Image, lower: float = 1, upper: float = 2
) -> Image.Image:
    """Enhance the image's contrast by a random factor.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random enhancement.
        Default is 0.5.
    :param upper: Upper bound for the random enhancement.
        Default is 1.5.
    :return: Adjusted image.
    """
    enhancer = ImageEnhance.Contrast(img)
    return enhancer.enhance(random.uniform(lower, upper))


def decrease_contrast(
    img: Image.Image, lower: float = 0.5, upper: float = 1
) -> Image.Image:
    """Reduce the image's contrast by a random factor.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random enhancement.
        Default is 0.5.
    :param upper: Upper bound for the random enhancement.
        Default is 1.5.
    :return: Adjusted image.
    """
    enhancer = ImageEnhance.Contrast(img)
    return enhancer.enhance(random.uniform(lower, upper))


def add_saturation(
    img: Image.Image, lower: float = 1, upper: float = 2
) -> Image.Image:
    """Enhance the image's color saturation by a random factor.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random enhancement.
        Default is 0.5.
    :param upper: Upper bound for the random enhancement.
        Default is 1.5.
    :return: Adjusted image.
    """
    enhancer = ImageEnhance.Color(img)
    return enhancer.enhance(random.uniform(lower, upper))


def add_brightness(
    img: Image.Image, lower: float = 1, upper: float = 2
) -> Image.Image:
    """Increase the image's brightness by a random factor.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random enhancement.
        Default is 0.5.
    :param upper: Upper bound for the random enhancement.
        Default is 1.5.
    :return: Adjusted image.
    """
    enhancer = ImageEnhance.Brightness(img)
    return enhancer.enhance(random.uniform(lower, upper))


def add_darkness(
    img: Image.Image, lower: float = 0.5, upper: float = 1
) -> Image.Image:
    """Decrease the image's brightness by a random factor.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random enhancement.
        Default is 0.5.
    :param upper: Upper bound for the random enhancement.
        Default is 1.5.
    :return: Adjusted image.
    """
    enhancer = ImageEnhance.Brightness(img)
    return enhancer.enhance(random.uniform(lower, upper))


def flip_vertical(img: Image.Image) -> Image.Image:
    """Flip the image vertically."""
    return img.transpose(method=Image.FLIP_TOP_BOTTOM)


def flip_horizontal(img: Image.Image) -> Image.Image:
    """Flip the image horizontally."""
    return img.transpose(method=Image.FLIP_LEFT_RIGHT)


def rotate(
    img: Image.Image,
    lower: int = -45,
    upper: int = 45,
) -> Image.Image:
    """Rotate the image by a random angle.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random rotation.
        Default is 0.5.
    :param upper: Upper bound for the random rotation.
        Default is 1.5.
    :return: Adjusted image.
    """
    angle = random.randint(lower, upper)
    return img.rotate(angle)


def gaussian_blur(
    img: Image.Image, lower: float = 0.5, upper: float = 3
) -> Image.Image:
    """Apply Gaussian blur to the image using a random radius.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random radius.
        Default is 0.5.
    :param upper: Upper bound for the random radius.
        Default is 1.5.
    :return: Adjusted image.
    """
    return img.filter(
        ImageFilter.GaussianBlur(radius=random.uniform(lower, upper))
    )


def solarize(img: Image.Image, threshold: int = 128) -> Image.Image:
    """Invert pixel values above a specified threshold."""
    if img.mode == "RGBA":
        # Split the image into RGB and alpha
        rgb, alpha = img.split()[0:3], img.split()[3]

        # Convert the RGB tuple back to an image
        rgb_img = Image.merge("RGB", rgb)

        # Solarize the RGB image
        solarized_rgb = ImageOps.solarize(rgb_img, threshold=threshold)

        # Merge back with the alpha channel
        solarized_img = Image.merge("RGBA", (*solarized_rgb.split(), alpha))

        return solarized_img
    else:
        return ImageOps.solarize(img, threshold=threshold)


def random_crop(
    img: Image.Image, lower: float = 0.6, upper: float = 0.9
) -> Image.Image:
    """Randomly crop a portion of the image.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random crop.
        Default is 0.5.
    :param upper: Upper bound for the random crop.
        Default is 1.5.
    :return: Adjusted image.
    """
    width, height = img.size
    crop_size = random.uniform(lower, upper)
    new_width, new_height = int(width * crop_size), int(height * crop_size)

    left = random.randint(0, width - new_width)
    top = random.randint(0, height - new_height)
    right = left + new_width
    bottom = top + new_height

    return img.crop((left, top, right, bottom))


def equalize(img: Image.Image) -> Image.Image:
    """Equalize the image's histogram."""
    return ImageOps.equalize(img)


def color_jitter(
    img: Image.Image, lower: float = 0.5, upper: float = 1.5
) -> Image.Image:
    """Randomly adjust the image's brightness, contrast, saturation, and hue.

    :param img: Input image to be adjusted.
    :param lower: Lower bound for the random enhancement multiplier.
        Default is 0.5.
    :param upper: Upper bound for the random enhancement multiplier.
        Default is 1.5.
    :return: Adjusted image.
    """
    img = ImageEnhance.Brightness(img).enhance(random.uniform(lower, upper))
    img = ImageEnhance.Contrast(img).enhance(random.uniform(lower, upper))
    img = ImageEnhance.Color(img).enhance(random.uniform(lower, upper))
    return img


DEFAULT_AUGMENTATIONS: List[Tuple[Callable, Dict[str, Any]]] = [
    (add_brightness, {}),
    (add_contrast, {}),
    (add_darkness, {}),
    (add_saturation, {}),
    (decrease_contrast, {}),
    (flip_horizontal, {}),
    (flip_vertical, {}),
    (grayscale, {}),
    (resize_height, {}),
    (resize_width, {}),
    (rotate, {}),
    (solarize, {}),
    (gaussian_blur, {}),
    # Additional
    # (equalize, {}),
    # (color_jitter, {}),
    # (random_crop, {}),
]


def augment_image(
    image_bytes: bytes,
    augmentations: Optional[List[Tuple[Callable, Dict[str, Any]]]] = None,
    num_steps: Optional[int] = None,
    pop_func: Callable = random_pop,
) -> bytes:
    """Augment the input image with a series of random augmentation methods.

    Read an image provided in bytes format, applies a specified number
    of random augmentation methods from a given list, and then returns the
    augmented image in bytes format. If no list of methods is provided, a
    default list is used. If no number of steps (methods) is specified, all
    methods will be applied.

    :param image_bytes: Input image in bytes format.
    :param augmentations: List of tuples of callable augmentation
        functions and their respective keyword arguments. If not
        provided, the default augmentation functions will be used.
    :param num_steps: Number of augmentation steps (functions) to be
        applied. If not specified, the length of the `augmentations` list
        will be used.
    :param pop_func: Callable to pop items from `augmentations` list. By
        default, the `random_pop` is used, which pops items in random
        order. If you want the order of augmentations to be constant and
        as given, replace it with `list.pop` (`pop_func=list.pop`).
    :return: Augmented image in bytes format.
    """
    # Load the image using PIL
    image = Image.open(io.BytesIO(image_bytes))
    # Original file format somehow gets lots during conversion.
    # We save it for later.
    image_format = image.format

    # Convert to RGB if the image is in palette mode. If we don't do so,
    # some formats (namely GIF) will certainly break on this.
    if image.mode == "P":
        image = image.convert("RGB")

    counter = 0

    if not augmentations:
        _augmentations = deepcopy(DEFAULT_AUGMENTATIONS)
    else:
        _augmentations = deepcopy(augmentations)

    if not num_steps:
        num_steps = len(_augmentations)

    while counter < num_steps:
        func_and_kwargs = pop_func(_augmentations)
        if func_and_kwargs:
            func, kwargs = func_and_kwargs
            # LOGGER.debug(f"Applying {func} to {id(image_bytes)}")
            try:
                image = func(image, **kwargs)
            except Exception as err:
                # Some combination of filters may not work correctly together.
                # Therefore, we silence the errors here.
                LOGGER.warning(f"Failed to apply {func} to {id(image_bytes)}")
                LOGGER.exception(err)
            counter += 1

    # Convert the image back to bytes
    byte_array = io.BytesIO()
    image.save(byte_array, format=image_format)
    return byte_array.getvalue()


def augment_image_file(
    image_path: str,
    augmentations: Optional[List[Tuple[Callable, Dict[str, Any]]]] = None,
    num_steps: Optional[int] = None,
    pop_func: Callable = random_pop,
) -> bytes:
    """Augment image from path.

    Augment the input image with a series of random augmentation functions.
    """
    with open(image_path, "rb") as image_bytes:
        return augment_image(
            image_bytes=image_bytes.read(),
            augmentations=augmentations,
            num_steps=num_steps,
            pop_func=pop_func,
        )



from faker import Faker
from faker_file.providers.image.pil_generator import PilImageGenerator
from faker_file.providers.png_file import PngFileProvider

FAKER = Faker()
FAKER.add_provider(PngFileProvider)

png_file = FAKER.png_file(
    image_generator_cls=PilImageGenerator,
    image_generator_kwargs={
        "encoding": "utf8",
        "font_size": 14,
        "page_width": 800,
        "page_height": 1200,
        "line_height": 16,
        "spacing": 5,
    },
    wrap_chars_after=100,
)



import logging
import unittest
from copy import deepcopy
from typing import Any, Dict, List, Tuple, Type, Union

import pytest
from faker import Faker
from fuzzywuzzy import fuzz
from parametrize import parametrize
from tika import parser

from ..providers.docx_file import DocxFileProvider
from ..providers.eml_file import EmlFileProvider
from ..providers.epub_file import EpubFileProvider
from ..providers.odp_file import OdpFileProvider
from ..providers.ods_file import OdsFileProvider
from ..providers.odt_file import OdtFileProvider
from ..providers.pdf_file import PdfFileProvider
from ..providers.pdf_file.generators.pdfkit_generator import PdfkitPdfGenerator
from ..providers.pdf_file.generators.reportlab_generator import (
    ReportlabPdfGenerator,
)
from ..providers.pptx_file import PptxFileProvider
from ..providers.rtf_file import RtfFileProvider
from ..providers.txt_file import TxtFileProvider
from ..providers.xlsx_file import XlsxFileProvider
from ..providers.xml_file import XmlFileProvider
from ..registry import FILE_REGISTRY

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("DataIntegrityTestCase",)

FAKER = Faker()
FAKER.add_provider(DocxFileProvider)
FAKER.add_provider(EmlFileProvider)
FAKER.add_provider(EpubFileProvider)
FAKER.add_provider(OdpFileProvider)
FAKER.add_provider(OdsFileProvider)
FAKER.add_provider(OdtFileProvider)
FAKER.add_provider(PdfFileProvider)
FAKER.add_provider(PptxFileProvider)
FAKER.add_provider(RtfFileProvider)
FAKER.add_provider(TxtFileProvider)
FAKER.add_provider(XlsxFileProvider)
FAKER.add_provider(XmlFileProvider)

FileProvider = Union[
    DocxFileProvider,
    EmlFileProvider,
    EpubFileProvider,
    OdpFileProvider,
    OdsFileProvider,
    OdtFileProvider,
    PdfFileProvider,
    PptxFileProvider,
    RtfFileProvider,
    TxtFileProvider,
    XlsxFileProvider,
    XmlFileProvider,
]

LOGGER = logging.getLogger(__name__)


class DataIntegrityTestCase(unittest.TestCase):
    """Data integrity tests.

    Methodology:

    - Generate file using given provider.
    - Extract text using Apache Tika.
    - Compare original text to the extracted text. If match is less
      than the specified margin, test fails.
    """

    # fake, provider, method_name, kwargs, similarity_margin
    __PARAMETRIZED_DATA: List[
        Tuple[
            Faker,
            Type[FileProvider],
            str,
            Dict[str, Any],
            int,
        ]
    ] = [
        # DOCX
        (FAKER, DocxFileProvider, "docx_file", {"max_nb_chars": 2_048}, 95),
        # EML
        (FAKER, EmlFileProvider, "eml_file", {"max_nb_chars": 2_048}, 95),
        # EPUB
        (FAKER, EpubFileProvider, "epub_file", {"max_nb_chars": 2_048}, 90),
        # ODP
        (FAKER, OdpFileProvider, "odp_file", {"max_nb_chars": 2_048}, 100),
        # ODS
        (FAKER, OdsFileProvider, "ods_file", {}, 75),
        # ODT
        (FAKER, OdtFileProvider, "odt_file", {"max_nb_chars": 2_048}, 100),
        # PDF
        (
            FAKER,
            PdfFileProvider,
            "pdf_file",
            {
                "max_nb_chars": 2_048,
                "pdf_generator_cls": PdfkitPdfGenerator,
            },
            85,
        ),
        (
            FAKER,
            PdfFileProvider,
            "pdf_file",
            {
                "max_nb_chars": 2_048,
                "pdf_generator_cls": ReportlabPdfGenerator,
            },
            85,
        ),
        # PPTX
        (FAKER, PptxFileProvider, "pptx_file", {"max_nb_chars": 2_048}, 95),
        # RTF
        (FAKER, RtfFileProvider, "rtf_file", {"max_nb_chars": 2_048}, 100),
        # TXT
        (FAKER, TxtFileProvider, "txt_file", {"max_nb_chars": 2_048}, 100),
        # XLSX
        (FAKER, XlsxFileProvider, "xlsx_file", {}, 75),
        # XML
        (FAKER, XmlFileProvider, "xml_file", {}, 95),
    ]

    def tearDown(self) -> None:
        super().tearDown()
        FILE_REGISTRY.clean_up()

    @parametrize(
        "fake, provider, method_name, kwargs, similarity_margin",
        __PARAMETRIZED_DATA,
    )
    @pytest.mark.flaky(reruns=3)
    def test_file_content(
        self: "DataIntegrityTestCase",
        fake: Faker,
        provider: Type[FileProvider],
        method_name: str,
        kwargs: Dict[str, Any],
        similarity_margin: int,
    ):
        _method = getattr(fake, method_name)
        _kwargs = deepcopy(kwargs)

        # Generate a file
        _file = _method(**_kwargs)

        # Extract the content from the generated file using Apache Tika
        parsed = parser.from_file(_file.data["filename"])
        extracted_content = parsed["content"]

        # Compare the original file content to the extracted content after
        # cleaning up both (remove double white-spaces).
        similarity = fuzz.ratio(
            " ".join(_file.data["content"].split()),
            " ".join(extracted_content.split()),
        )

        LOGGER.info(f"Provider: {method_name}, Similarity: {similarity}")
        if similarity < 100:
            LOGGER.warning(
                f"Original content: {_file.data['content']}\n"
                f"Extracted content: {extracted_content}"
            )

        # Assert that the similarity is at least the required margin %
        self.assertGreaterEqual(
            similarity,
            similarity_margin,
            f"Content of the generated {method_name} matches "
            f"the extracted content only for {similarity}% while "
            f"{similarity_margin}% is expected. \n"
            f"Original content: {_file.data['content']}\n"
            f"Extracted content: {extracted_content}",
        )



from faker import Faker
from faker_file.providers.eml_file import EmlFileProvider
from faker_file.providers.helpers.inner import create_inner_docx_file

FAKER = Faker()
FAKER.add_provider(EmlFileProvider)

eml_file = FAKER.eml_file(
    prefix="zzz",
    options={
        "count": 3,
        "create_inner_file_func": create_inner_docx_file,
        "create_inner_file_args": {
            "prefix": "xxx_",
            "max_nb_chars": 1_024,
        },
    },
)



import logging
import re
import subprocess
import unittest
from importlib import import_module, reload
from typing import Union

from parametrize import parametrize

from ..cli.command import main
from ..registry import FILE_REGISTRY
from ..storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("TestCLI",)

LOGGER = logging.getLogger(__name__)


VERSION_PATTERN = re.compile(r"^\d+(\.\d+){0,2}$")
FS_STORAGE = FileSystemStorage()
FILE_PATH_PATTERN = re.compile(r"/.+")


def convert_value_to_cli_arg(value) -> str:
    if isinstance(value, bool):
        return str(value).lower()
    elif isinstance(value, (int, float, str)):
        return str(value)
    else:
        raise ValueError(f"Unsupported value type: {type(value)}")


def extract_filename(val) -> Union[str, None]:
    match = FILE_PATH_PATTERN.search(val)
    if match:
        extracted_path = match.group(0)
        return extracted_path


class TestCLI(unittest.TestCase):
    """CLI tests."""

    def tearDown(self) -> None:
        super().tearDown()
        FILE_REGISTRY.clean_up()

    @parametrize(
        "method_name, kwargs",
        [
            # BIN
            (
                "bin_file",
                {},
            ),
            (
                "bin_file",
                {"length": 1_024},
            ),
            # CSV
            (
                "csv_file",
                {},
            ),
            (
                "csv_file",
                {"num_rows": 20},
            ),
            # DOCX
            (
                "docx_file",
                {},
            ),
            (
                "docx_file",
                {"wrap_chars_after": 40},
            ),
            # EML
            (
                "eml_file",
                {},
            ),
            (
                "eml_file",
                {"wrap_chars_after": 40},
            ),
            # EPUB
            (
                "epub_file",
                {},
            ),
            (
                "epub_file",
                {"wrap_chars_after": 40},
            ),
            # Generic
            (
                "generic_file",
                {
                    "content": "<html><body><p>{{text}}</p></body></html>",
                    "extension": "html",
                },
            ),
            # ICO
            (
                "ico_file",
                {},
            ),
            (
                "ico_file",
                {"wrap_chars_after": 40},
            ),
            # JPEG
            (
                "jpeg_file",
                {},
            ),
            (
                "jpeg_file",
                {"wrap_chars_after": 40},
            ),
            # MP3
            (
                "mp3_file",
                {},
            ),
            (
                "mp3_file",
                {"max_nb_chars": 256},
            ),
            # ODP
            (
                "odp_file",
                {},
            ),
            (
                "odp_file",
                {"wrap_chars_after": 40},
            ),
            # ODS
            (
                "ods_file",
                {},
            ),
            (
                "ods_file",
                {"num_rows": 20},
            ),
            # ODT
            (
                "odt_file",
                {},
            ),
            (
                "odt_file",
                {"wrap_chars_after": 40},
            ),
            # PDF
            (
                "pdf_file",
                {},
            ),
            (
                "pdf_file",
                {"wrap_chars_after": 40},
            ),
            # PNG
            (
                "png_file",
                {},
            ),
            (
                "png_file",
                {"wrap_chars_after": 40},
            ),
            # PPTX
            (
                "pptx_file",
                {},
            ),
            (
                "pptx_file",
                {"wrap_chars_after": 40},
            ),
            # RTF
            (
                "rtf_file",
                {},
            ),
            (
                "rtf_file",
                {"wrap_chars_after": 40},
            ),
            # SVG
            (
                "svg_file",
                {},
            ),
            (
                "svg_file",
                {"wrap_chars_after": 40},
            ),
            # TAR
            (
                "tar_file",
                {},
            ),
            (
                "tar_file",
                {"prefix": "ttt_"},
            ),
            # TXT
            (
                "txt_file",
                {},
            ),
            (
                "txt_file",
                {"wrap_chars_after": 40},
            ),
            # # WEBP
            # (
            #     "webp_file",
            #     {},
            # ),
            # (
            #     "webp_file",
            #     {},
            # ),
            # XLSX
            (
                "xlsx_file",
                {},
            ),
            (
                "xlsx_file",
                {"num_rows": 20},
            ),
            # ZIP
            (
                "zip_file",
                {},
            ),
            (
                "zip_file",
                {"prefix": "ttt_"},
            ),
        ],
    )
    def test_cli(self: "TestCLI", method_name: str, kwargs: dict) -> None:
        """Test CLI."""
        # Convert kwargs to command-line arguments
        args = [
            f"--{key}={convert_value_to_cli_arg(value)}"
            for key, value in kwargs.items()
        ]

        # Merge the base command with the generated arguments
        cmd = ["faker-file", method_name] + args
        # LOGGER.debug(f"cmd: {cmd}")
        # Execute the command with the provided arguments
        res = subprocess.check_output(cmd).strip()

        # Extract the filename to verify existence and clean-up
        filename = extract_filename(res.decode())
        # LOGGER.debug(f"filename: {filename}")
        self.assertTrue(filename)
        self.assertTrue(FS_STORAGE.exists(filename))
        FS_STORAGE.unlink(filename)

    def test_cli_error_no_provider(self: "TestCLI") -> None:
        """Test CLI, no provider given."""
        with self.assertRaises(subprocess.CalledProcessError):
            res = subprocess.check_output(["faker-file"]).strip()
            self.assertFalse(res)

    def test_cli_generate_completion(self: "TestCLI") -> None:
        """Test CLI, generate-completion."""
        cmd = ["faker-file", "generate-completion"]
        res = subprocess.check_output(cmd).strip()
        self.assertTrue(res)

    def test_cli_version(self: "TestCLI") -> None:
        """Test CLI, version."""
        cmd = ["faker-file", "version"]
        res = subprocess.check_output(cmd).strip()
        self.assertTrue(VERSION_PATTERN.match(res.decode()))

    def test_broken_imports(self: "TestCLI") -> None:
        """Test broken imports."""
        _module = import_module("faker_file.cli.helpers")
        del _module.__dict__["PROVIDERS"]
        with self.assertRaises(SystemExit):
            main()
        reload(_module)



from typing import Any, Dict, Union

from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("BaseImageGenerator",)


class BaseImageGenerator:
    """Base image generator."""

    def __init__(
        self: "BaseImageGenerator",
        **kwargs,
    ) -> None:
        """Constructor.

        :param kwargs: Dictionary with parameters (for PDF generator
            specific tuning).
        """
        self.handle_kwargs(**kwargs)

    def handle_kwargs(self: "BaseImageGenerator", **kwargs):
        """Handle kwargs."""

    def generate(
        self: "BaseImageGenerator",
        content: str,
        data: Dict[str, Any],
        provider: Union[Faker, Generator, Provider],
    ) -> bytes:
        raise NotImplementedError("Method `generate` is not implemented.")



"""
.. code-block:: python

    from faker import Faker
    from faker_file.base import DynamicTemplate
    from faker_file.contrib.pdf_file.pil_snippets import *
    from faker_file.providers.pdf_file import PdfFileProvider
    from faker_file.providers.pdf_file.generators.pil_generator import (
        PilPdfGenerator
    )

    FAKER = Faker()
    FAKER.add_provider(PdfFileProvider)

    file = FAKER.pdf_file(
        pdf_generator_cls=PilPdfGenerator,
        content=DynamicTemplate(
            [
                (add_h1_heading, {}),
                (add_paragraph, {"max_nb_chars": 500}),
                (add_paragraph, {"max_nb_chars": 500}),
                (add_paragraph, {"max_nb_chars": 500}),
                (add_paragraph, {"max_nb_chars": 500}),
            ]
        )
    )

    file = FAKER.pdf_file(
        pdf_generator_cls=PilPdfGenerator,
        content=DynamicTemplate(
            [
                (add_h1_heading, {}),
                (add_paragraph, {}),
                (add_picture, {}),
                (add_paragraph, {}),
                (add_picture, {}),
                (add_paragraph, {}),
                (add_picture, {}),
                (add_paragraph, {}),
            ]
        )
    )

    file = FAKER.pdf_file(
        pdf_generator_cls=PilPdfGenerator,
        content=DynamicTemplate(
            [
                (add_h1_heading, {}),
                (add_picture, {}),
                (add_paragraph, {"max_nb_chars": 500}),
                (add_picture, {}),
                (add_paragraph, {"max_nb_chars": 500}),
                (add_picture, {}),
                (add_paragraph, {"max_nb_chars": 500}),
                (add_picture, {}),
                (add_paragraph, {"max_nb_chars": 500}),
            ]
        )
    )

    file = FAKER.pdf_file(
        pdf_generator_cls=PilPdfGenerator,
        content=DynamicTemplate(
            [
                (add_h1_heading, {}),
                (add_picture, {}),
                (add_paragraph, {"max_nb_chars": 500}),
                (add_table, {"rows": 5, "cols": 4}),
            ]
        )
    )

    file = FAKER.pdf_file(
        pdf_generator_cls=PilPdfGenerator,
        content=DynamicTemplate(
            [
                (add_h1_heading, {"margin": (2, 2)}),
                (add_picture, {"margin": (2, 2)}),
                (add_paragraph, {"max_nb_chars": 500, "margin": (2, 2)}),
                (add_picture, {"margin": (2, 2)}),
                (add_paragraph, {"max_nb_chars": 500, "margin": (2, 2)}),
                (add_picture, {"margin": (2, 2)}),
                (add_paragraph, {"max_nb_chars": 500, "margin": (2, 2)}),
                (add_picture, {"margin": (2, 2)}),
                (add_paragraph, {"max_nb_chars": 500, "margin": (2, 2)}),
            ]
        )
    )
"""

import logging
import textwrap
from collections import namedtuple
from io import BytesIO
from typing import Tuple, Union

from PIL import Image, ImageFont

from ...base import DEFAULT_FORMAT_FUNC

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "add_h1_heading",
    "add_h2_heading",
    "add_h3_heading",
    "add_h4_heading",
    "add_h5_heading",
    "add_h6_heading",
    "add_heading",
    "add_page_break",
    "add_paragraph",
    "add_picture",
    "add_table",
)

LOGGER = logging.getLogger(__name__)


def expand_margin(
    margin: Union[Tuple[int, int], Tuple[int, int, int, int]]
) -> Tuple[int, int, int, int]:
    """Utility function to expand the margin tuple."""
    if len(margin) == 2:
        top, right = margin
        return top, right, top, right
    elif len(margin) == 4:
        return margin


def add_picture(
    provider,
    generator,
    data,
    counter,
    **kwargs,
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for picture generation using PIL."""
    # Extract margin values
    margin = kwargs.get("margin", (0, 0))
    top_margin, right_margin, bottom_margin, left_margin = expand_margin(margin)

    image_bytes = kwargs.get(
        "image_bytes", provider.generator.image()
    )  # Assuming image() returns bytes
    # Create a BytesIO object and load the image data
    with BytesIO(image_bytes) as input_stream:
        pil_image = Image.open(input_stream)

        # Resize the image
        new_width = kwargs.get("image_width", 200)
        new_height = kwargs.get("image_height", 200)
        if new_width > generator.page_width:
            new_width = generator.page_width
        if new_height > generator.page_height:
            new_height = generator.page_height
        pil_image = pil_image.resize((new_width, new_height))

        # Create a BytesIO object outside the 'with' statement
        output_stream = BytesIO()
        pil_image.save(output_stream, format="PNG")
        output_stream.seek(0)  # Move to the start of the stream

    # X, Y coordinates where the text will be placed
    # position = kwargs.get("position", (0, 0))
    # Adjust the position with margin for the left and top
    position = (
        kwargs.get("position", (0, 0))[0] + left_margin,
        kwargs.get("position", (0, 0))[1] + top_margin,
    )

    # Calculate the remaining space on the current page
    remaining_space = generator.page_height - position[1]

    # Create a PIL Image object from bytes
    image_to_paste = Image.open(output_stream)

    # Check if the image will fit on the current page
    # LOGGER.debug(f"image_to_paste.height: {image_to_paste.height}")
    # LOGGER.debug(f"remaining_space: {remaining_space}")
    if remaining_space < image_to_paste.height:
        # Image won't fit; add the current page to the list and create a new one
        generator.save_and_start_new_page()

        # Reset position to start of new page
        position = (0, 0)

    # Ensure that the document and the image to paste have the same mode
    if generator.img.mode != image_to_paste.mode:
        image_to_paste = image_to_paste.convert(generator.img.mode)

    # Create a mask if the image has an alpha channel
    mask = None
    if "A" in image_to_paste.getbands():
        mask = image_to_paste.split()[3]

    # Paste the image into the document
    generator.img.paste(image_to_paste, position, mask)

    # If you want to keep track of the last position to place
    # another element, you can.
    # last_position = (
    #     position[0] + image.width, position[1] + image_to_paste.height
    # )
    last_position = (0, position[1] + image_to_paste.height + bottom_margin)

    # Meta-data (optional)
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_picture", {})
    data["content_modifiers"]["add_picture"].setdefault(counter, [])
    data["content_modifiers"]["add_picture"][counter].append("Image added")

    return False, last_position


def add_paragraph(
    provider,
    generator,
    data,
    counter,
    **kwargs,
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for paragraph generation using PIL."""
    # Extract margin values
    margin = kwargs.get("margin", (0, 0))
    top_margin, right_margin, bottom_margin, left_margin = expand_margin(margin)
    content = kwargs.get("content", None)
    max_nb_chars = kwargs.get("max_nb_chars", 5_000)
    wrap_chars_after = kwargs.get("wrap_chars_after", None)
    # X, Y coordinates where the text will be placed
    # position = kwargs.get("position", (0, 0))
    # position = kwargs.get("position", (0, 0))
    # Adjust the position with margin for the left and top
    position = (
        kwargs.get("position", (0, 0))[0] + left_margin,
        kwargs.get("position", (0, 0))[1] + top_margin,
    )
    content_specs = kwargs.get("content_specs", {})
    format_func = kwargs.get("format_func", DEFAULT_FORMAT_FUNC)

    _content = provider._generate_text_content(
        max_nb_chars=max_nb_chars,
        wrap_chars_after=wrap_chars_after,
        content=content,
        format_func=format_func,
    )
    font = ImageFont.truetype(generator.font, generator.font_size)
    lines = _content.split("\n")
    line_max_num_chars = generator.find_max_fit_for_multi_line_text(
        generator.draw,
        lines,
        font,
        generator.page_width,
    )
    wrap_chars_after = content_specs.get("wrap_chars_after")
    if (
        not wrap_chars_after
        or wrap_chars_after
        and (wrap_chars_after > line_max_num_chars)
    ):
        lines = textwrap.wrap(_content, line_max_num_chars)

    # Load a truetype or opentype font file, and create a font object.
    font = ImageFont.truetype(generator.font, generator.font_size)

    y_text = position[1]
    # LOGGER.debug(f"position: {position}")
    for counter, line in enumerate(lines):
        text_width, text_height = generator.draw.textsize(
            line, font=font, spacing=generator.spacing
        )
        if y_text + text_height > generator.page_height:
            generator.save_and_start_new_page()
            y_text = 0

        generator.draw.text(
            (position[0], y_text),
            line,
            fill=(0, 0, 0),
            spacing=generator.spacing,
            font=font,
        )
        # Move down for next line
        y_text += text_height + generator.line_height

    # If you want to keep track of the last position to place another
    # element, you can.
    # last_position = (position[0], y_text)
    # last_position = (0, y_text)
    last_position = (0, y_text + bottom_margin)

    # Add meta-data, assuming data is a dictionary for tracking
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_paragraph", {})
    data["content_modifiers"]["add_paragraph"].setdefault(counter, [])
    data["content_modifiers"]["add_paragraph"][counter].append(_content)
    data.setdefault("content", "")
    data["content"] += "\r\n" + _content

    return False, last_position


def add_page_break(
    provider,
    generator,
    data,
    counter,
    **kwargs,
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for paragraph generation using PIL."""
    generator.save_and_start_new_page()

    # If you want to keep track of the last position to place another
    # element, you can.
    # last_position = (position[0], 0)
    last_position = (0, 0)

    return False, last_position


def get_heading_font_size(base_size: int, heading_level: int) -> int:
    return base_size * (8 - heading_level) // 2


def add_heading(
    provider,
    generator,
    data,
    counter,
    **kwargs,
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for H1 heading generation using PIL."""
    # Extract margin values
    margin = kwargs.get("margin", (0, 0))
    top_margin, right_margin, bottom_margin, left_margin = expand_margin(margin)
    content = kwargs.get("content", None)
    max_nb_chars = kwargs.get("max_nb_chars", 30)
    wrap_chars_after = kwargs.get("wrap_chars_after", None)
    format_func = kwargs.get("format_func", DEFAULT_FORMAT_FUNC)
    # X, Y coordinates where the text will be placed
    # position = kwargs.get("position", (0, 0))
    # Adjust the position with margin for the left and top
    position = (
        kwargs.get("position", (0, 0))[0] + left_margin,
        kwargs.get("position", (0, 0))[1] + top_margin,
    )
    level = kwargs.get("level", 1)
    if level < 1 or level > 6:
        level = 1

    font_size = get_heading_font_size(generator.font_size, level)

    _content = provider._generate_text_content(
        max_nb_chars=max_nb_chars,
        wrap_chars_after=wrap_chars_after,
        content=content,
        format_func=format_func,
    )

    # Here, you'll specify a different font size for heading
    font = ImageFont.truetype(generator.font, font_size)

    y = position[1]
    generator.draw.text(
        (position[0], y),
        _content,
        fill=(0, 0, 0),
        font=font,
    )

    text_width, text_height = generator.draw.textsize(_content, font=font)
    y += text_height

    # If you want to keep track of the last position to place another
    # element, you can.
    # last_position = (position[0], y)
    # last_position = (0, y)
    last_position = (0, y + bottom_margin)

    # Add meta-data, assuming data is a dictionary for tracking
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_heading", {})
    data["content_modifiers"]["add_heading"].setdefault(counter, [])
    data["content_modifiers"]["add_heading"][counter].append(_content)
    data.setdefault("content", "")
    data["content"] += "\r\n" + _content

    return False, last_position


def add_h1_heading(
    provider, generator, data, counter, **kwargs
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for the h1 heading generation."""
    return add_heading(provider, generator, data, counter, level=1, **kwargs)


def add_h2_heading(
    provider, generator, data, counter, **kwargs
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for the h2 heading generation."""
    return add_heading(provider, generator, data, counter, level=2, **kwargs)


def add_h3_heading(
    provider, generator, data, counter, **kwargs
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for the h3 heading generation."""
    return add_heading(provider, generator, data, counter, level=3, **kwargs)


def add_h4_heading(
    provider, generator, data, counter, **kwargs
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for the h4 heading generation."""
    return add_heading(provider, generator, data, counter, level=4, **kwargs)


def add_h5_heading(
    provider, generator, data, counter, **kwargs
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for the h5 heading generation."""
    return add_heading(provider, generator, data, counter, level=5, **kwargs)


def add_h6_heading(
    provider, generator, data, counter, **kwargs
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for the h6 heading generation."""
    return add_heading(provider, generator, data, counter, level=6, **kwargs)


# This is a simple placeholder for your table object
Table = namedtuple("Table", ["data"])


def draw_table_cell(document, cell_content, position, cell_size, font):
    """Draw a table cell."""
    x, y = position
    width, height = cell_size
    border_color = (0, 0, 0)
    fill_color = (255, 255, 255)

    # Draw the rectangle
    document.rectangle(
        [position, (x + width, y + height)],
        fill=fill_color,
        outline=border_color,
    )

    # Draw text in the rectangle
    text_width, text_height = document.textsize(cell_content, font=font)
    text_position = (
        x + (width - text_width) // 2,
        y + (height - text_height) // 2,
    )
    document.text(text_position, cell_content, fill=(0, 0, 0), font=font)


def add_table(
    provider,
    generator,
    data,
    counter,
    **kwargs,
) -> Tuple[bool, Tuple[int, int]]:
    """Callable responsible for table generation using PIL."""
    # X, Y coordinates where the table will be placed
    position = kwargs.get("position", (0, 0))

    # Default cell dimensions
    cell_width = kwargs.get("cell_width", 100)
    cell_height = kwargs.get("cell_height", 30)

    # Font for the table cells
    font = ImageFont.truetype(generator.font, generator.font_size)

    # Extract or generate table data
    rows = kwargs.get("rows", 3)
    cols = kwargs.get("cols", 4)
    headers = [f"Header {i + 1}" for i in range(cols)]
    table_data = [
        [provider.generator.word() for _ in range(cols)] for _ in range(rows)
    ]
    table_data.insert(0, headers)
    table = Table(table_data)

    y = position[1]
    for row in table.data:
        x = position[0]
        for cell_content in row:
            cell_position = (x, y)
            draw_table_cell(
                generator.draw,
                cell_content,
                cell_position,
                (cell_width, cell_height),
                font,
            )
            x += cell_width
        y += cell_height

    last_position = (0, y)

    # Add meta-data, assuming data is a dictionary for tracking
    data.setdefault("content_modifiers", {})
    data["content_modifiers"].setdefault("add_table", {})
    data["content_modifiers"]["add_table"].setdefault(counter, [])
    data["content_modifiers"]["add_table"][counter].append("Table added")

    return False, last_position



import logging
from io import BytesIO
from typing import Any, Dict, Union

from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider
from reportlab.lib.pagesizes import letter
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.units import inch
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.platypus import Paragraph, SimpleDocTemplate

from ....base import DynamicTemplate
from ....constants import DEFAULT_FONT_NAME, DEFAULT_FONT_PATH
from ...base.pdf_generator import BasePdfGenerator

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("ReportlabPdfGenerator",)

LOGGER = logging.getLogger(__name__)


class ReportlabPdfGenerator(BasePdfGenerator):
    """Reportlab PDF generator.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.pdf_file import PdfFileProvider
        from faker_file.providers.pdf_file.generators import reportlab_generator

        FAKER = Faker()
        FAKER.add_provider(PdfFileProvider)

        file = FAKER.pdf_file(
            pdf_generator_cls=reportlab_generator.ReportlabPdfGenerator
        )

    Using `DynamicTemplate`:

    .. code-block:: python

        from io import BytesIO

        from faker import Faker
        from faker_file.base import DynamicTemplate
        from faker_file.providers.jpeg_file import JpegFileProvider
        from faker_file.providers.pdf_file import PdfFileProvider
        from faker_file.providers.pdf_file.generators import (
            reportlab_generator
        )
        from PIL import Image as PilImage
        from reportlab.lib.pagesizes import letter, landscape
        from reportlab.lib.styles import getSampleStyleSheet
        from reportlab.lib import colors
        from reportlab.pdfbase import pdfmetrics
        from reportlab.pdfbase.ttfonts import TTFont
        from reportlab.platypus import (
            Image,
            Table,
            TableStyle,
            PageBreak,
            Paragraph,
        )

        FAKER = Faker()
        FAKER.add_provider(PdfFileProvider)

        # Add table function
        def pdf_add_table(
            provider, generator, story, data, counter, **kwargs
        ):
            rows = kwargs.get("rows", 3)
            cols = kwargs.get("cols", 4)

            # Define your table headers
            headers = [f"Header {i+1}" for i in range(cols)]

            # Generate the rest of the table data
            table_data = [
                [
                    provider.generator.word() for _ in range(cols)
                ] for _ in range(rows)
            ]

            # Add the headers to the table data
            table_data.insert(0, headers)

            # Create the table object
            table = Table(table_data)

            # Apply table styles
            table.setStyle(
                TableStyle(
                    [
                        ("BACKGROUND", (0, 0), (-1, 0), colors.grey),
                        ("TEXTCOLOR", (0, 0), (-1, 0), colors.whitesmoke),
                        ("ALIGN", (0, 0), (-1, -1), "CENTER"),
                        ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                        ("FONTSIZE", (0, 0), (-1, 0), 14),
                        ("BOTTOMPADDING", (0, 0), (-1, 0), 12),
                        ("BACKGROUND", (0, 1), (-1, -1), colors.beige),
                        ("GRID", (0, 0), (-1, -1), 1, colors.black),
                        ("BOX", (0, 0), (-1, -1), 1, colors.black),
                    ]
                )
            )

            # Add the table to the document and build it
            story.append(table)

        # Add picture function
        def pdf_add_picture(
            provider, generator, story, data, counter, **kwargs
        ):
            jpeg_file = JpegFileProvider(provider.generator).jpeg_file(
                raw=True
            )

            # Create a BytesIO object and load the image data
            with BytesIO(jpeg_file) as input_stream:
                pil_image = PilImage.open(input_stream)

                # Resize the image
                new_width = 400
                new_height = 400
                pil_image = pil_image.resize((new_width, new_height))

                # Create a BytesIO object outside the 'with' statement
                output_stream = BytesIO()
                pil_image.save(output_stream, format='JPEG')
                output_stream.seek(0)  # Move to the start of the stream

                # Now you can use output_stream as your image data
                img = Image(output_stream)
                img.width = new_width
                img.height = new_height
                story.append(img)

        # Add page break function
        def pdf_add_page_break(
            provider, generator, story, data, counter, **kwargs
        ):
            # Insert a page break
            story.append(PageBreak())

        # Add paragraph function
        def pdf_add_paragraph(
            provider, generator, story, data, counter, **kwargs
        ):
            # Insert a paragraph
            styles = getSampleStyleSheet()
            style_paragraph = styles["Normal"]
            style_paragraph.fontName = generator.font_name
            pdfmetrics.registerFont(
                TTFont(generator.font_name, generator.font_path)
            )
            content = provider.generator.text(max_nb_chars=5_000)
            paragraph = Paragraph(content, style_paragraph)
            story.append(paragraph)

        # Create a file with table, page-break, picture, page-break, paragraph
        file = FAKER.pdf_file(
            pdf_generator_cls=(
                reportlab_generator.ReportlabPdfGenerator
            ),
            content=DynamicTemplate(
                [
                    (pdf_add_table, {}),
                    (pdf_add_page_break, {}),
                    (pdf_add_picture, {}),
                    (pdf_add_page_break, {}),
                    (pdf_add_paragraph, {}),
                ]
            )
        )

        # Create a file with text of 100 pages
        file = FAKER.pdf_file(
            pdf_generator_cls=(
                reportlab_generator.ReportlabPdfGenerator
            ),
            content=DynamicTemplate(
                [
                    (pdf_add_paragraph, {}),
                    (pdf_add_page_break, {}),
                ] * 100
            )
        )
    """

    font_name: str = DEFAULT_FONT_NAME
    font_path: str = DEFAULT_FONT_PATH

    def handle_kwargs(self: "ReportlabPdfGenerator", **kwargs) -> None:
        """Handle kwargs."""
        if "font_name" in kwargs:
            self.font_name = kwargs["font_name"]
        if "font_path" in kwargs:
            self.font_path = kwargs["font_path"]

    def generate(
        self: "ReportlabPdfGenerator",
        content: Union[str, DynamicTemplate],
        data: Dict[str, Any],
        provider: Union[Faker, Generator, Provider],
        **kwargs,
    ) -> bytes:
        """Generate PDF.

        :param content:
        :param data:
        :param provider: `PdfFileProvider` instance.
        """
        styles = getSampleStyleSheet()
        style_paragraph = styles["Normal"]
        style_paragraph.fontName = self.font_name
        pdfmetrics.registerFont(TTFont(self.font_name, self.font_path))

        story = []
        buffer = BytesIO()
        doc = SimpleDocTemplate(
            buffer,
            pagesize=letter,
            bottomMargin=0.4 * inch,
            topMargin=0.6 * inch,
            rightMargin=0.8 * inch,
            leftMargin=0.8 * inch,
        )
        if isinstance(content, DynamicTemplate):
            for counter, (ct_modifier, ct_modifier_kwargs) in enumerate(
                content.content_modifiers
            ):
                ct_modifier(
                    provider,
                    self,
                    story,
                    data,
                    counter,
                    **ct_modifier_kwargs,
                )
        else:
            paragraph = Paragraph(content, style_paragraph)
            story.append(paragraph)

        doc.build(story)

        return buffer.getvalue()



from faker import Faker

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "DOCX_KWARGS",
    "GCS_CREDENTIALS_JSON",
    "XML_DOWNLOAD_KWARGS",
    "XML_DOWNLOAD_TEMPLATE",
    "XML_ISBN_KWARGS",
    "XML_ISBN_TEMPLATE",
    "XML_METADATA_KWARGS",
    "XML_METADATA_TEMPLATE",
)

FAKER = Faker()
BOOK_ID = FAKER.pystr()

XML_METADATA_TEMPLATE = """
<book>
  <title>{{sentence}}</title>
  <description>{{paragraph}}</description>
  <date>{{date}}</date>
</book>
"""

XML_ISBN_TEMPLATE = """
<submission>
  <publisher id="{{pyint}}"/>
  <isbn10>{{isbn10}}</isbn10>
  <isbn13>{{isbn13}}</isbn13>
</submission>
"""

XML_DOWNLOAD_TEMPLATE = f"""
<download>
  <filegroup>
    <archive-file name="{BOOK_ID}.zip" />
    <metadata-file name="{BOOK_ID}_metadata.xml" />
    <file name="{BOOK_ID}.pdf" />
    <isbn-file name="{BOOK_ID}_isbn.xml" />
  </filegroup>
</download>
"""

XML_METADATA_KWARGS = {
    "content": XML_METADATA_TEMPLATE,
    "basename": f"{BOOK_ID}_metadata",
}
XML_DOWNLOAD_KWARGS = {
    "content": XML_DOWNLOAD_TEMPLATE,
    "basename": f"{BOOK_ID}_download",
}
XML_ISBN_KWARGS = {
    "content": XML_ISBN_TEMPLATE,
    "basename": f"{BOOK_ID}_isbn",
}
DOCX_KWARGS = {"basename": BOOK_ID}

GCS_CREDENTIALS_JSON = {
    "type": "service_account",
    "project_id": "test-12345",
    "private_key_id": "00000000000000000000000000",
    "private_key": "\n-----BEGIN PRIVATE KEY-----\nMIIEvgIBADANBgkqhkiG9w0BAQEFAASCBKgwggSkAgEAAoIBAQCfx5MW2anBbHKL\n4g1iazAtyxJGoHTfR03zGHr4mVsmfe2cV5vMZY4/oO3uszpbq342tdywfkAq5O8+\nFqbUUIEfAiobs5DVvZ9iL6uAkb/erLY2ObVFDr7hKBu0oRmc2EWYH6honxb/tNFI\nzLy3Sh7Bdor4/vH7PMalgcEorxt+RkFPJeo3cGyrJfO8jv0n2u0wakHKO5bPZshZ\nBAktnYDMVPnaV+kI/ezucxow91l3BXTdPuC3zVSlSuvqRssg8/jpDRgFv/hThuVm\nhUhU8ddS76QUQ0ekj6kBPp5iYvY6sJHwU9Bmf7qG3zAx4h56WBJV+SYDVoT+VydZ\nkqMc721LAgMBAAECggEAALPfxlDzH/U6VIa/W+ujjMzNnSvld60YZais5pAq4dkN\nHqV8yBuGONECjgzE6+3g7zXbhspdvXjzDskKUMXndQ7Z+o2bZgugbZq1QvdHhB/E\nT8OIzYkkEIw2Z8ckyRcv+/XfrermyGPi+IWV96rGxqXSNKG1lLV1HJM2gAqzfaZ1\nqL/TVxj9ZjI3FRGMSuxKeRC/RG6L4iHMq+tOIKSiST6UHSci6TbDDAWXpXMWzEeM\nmHEGCvNmp5yRHEPBF5n5Dl9hEfnhp8FpW20fuANjxIFlW9KMKtu9vkGgS1AdiT9f\nE1lgy41fBdwg2rdMHn3JhCN9H+J2MK7AtoWcD5TNGQKBgQDXszOluYO85hEbtJnN\nf5XtO/aoHtnjPixiI2wcxuGmLWaAYmtgy0KwWZlQldNXM7qH1BiEZ8wTFVOfqrqY\nlDuMYxgfTqPTaESS5wkirucMdZJJXuohLgDAaOaVbL0V+1YJbnq4AnLaHB0Y8sqX\nrnCawF1ke9F0032lbkQJn3qbJQKBgQC9obvTZO/53TrKz2/NWfBDJaTey7PaZ2fg\nJwvj63nUVVM91mQ262mjoqw/Wogp6GFr46XXcyQ/xk8g32EQ1a22p1MXIkdJTic3\n/m+VF//4+GaJD1MedqOxWMVWHZyYJxjekAM/tt3/epzcDFNdab6ZlRIyAh0Ro1mZ\nqJPmbCszrwKBgQCv1F2il2JDJswFaKgDcyCVHU9I5rU436KwcS2dG6Yvn0yyFQhx\nA+Ad/zvSDAAWUo2YUZWWwUICwFzFiBfJbvRH0TOFucYj/BgCJCE3S5n+dwzDkIKM\nf4KPVjO41MYiWBpfX9bbjutuzoINp0Tsdn9GNs8qrSAl+oyuwP7nVUBNnQKBgQCT\ne2fy/vvMnoyNE0vmr942ut5BELhuUiHtqTCMMKVtyHaXD1idhfWA+JFyLFzeCwdJ\nu6FNsRUuLHN6I4EAcM9L0VLEGTrL/mZuHAp4MFQ6NCa6zhpdBPRGh73iPeF+TFoB\nLov4T6bUfW3ljgiADC/ajp+6GP62qw6SfROaD+KBrQKBgHxz/IvyRE7zORNnhcld\n9y+Qr4GPLxV9oLeJlG0gjJaYC79gQk2gUjQfT2HEQlM76SeXr0qr/jOKSZeg4h7Q\nuZib/5m3YD5bI35qBexNHz/AZ1kQ9PiHjZmgS4PiCobuuHUXRPni0Vf0FKLCouQq\n0Bb9H46yp7+n+NCkjHB3IInM\n-----END PRIVATE KEY-----\n",  # noqa
    "client_email": "test-12345@appspot.gserviceaccount.com",
    "client_id": "2222222222222222222222222",
    "auth_uri": "https://accounts.google.com/o/oauth2/auth",
    "token_uri": "https://oauth2.googleapis.com/token",
    "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
    "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/test-12345%40appspot.gserviceaccount.com",  # noqa
    "universe_domain": "googleapis.com",
}



from faker import Faker
from faker_file.providers.ico_file import GraphicIcoFileProvider

FAKER = Faker()
FAKER.add_provider(GraphicIcoFileProvider)

ico_file = FAKER.graphic_ico_file(size=(800, 800))



import inspect
import os
import typing
from copy import deepcopy
from typing import Any, Dict, Tuple, Type

from faker import Faker

from ..base import FileMixin, StringValue
from ..providers.bin_file import BinFileProvider
from ..providers.csv_file import CsvFileProvider
from ..providers.docx_file import DocxFileProvider
from ..providers.eml_file import EmlFileProvider
from ..providers.epub_file import EpubFileProvider
from ..providers.generic_file import GenericFileProvider
from ..providers.ico_file import GraphicIcoFileProvider, IcoFileProvider
from ..providers.jpeg_file import GraphicJpegFileProvider, JpegFileProvider
from ..providers.mp3_file import Mp3FileProvider
from ..providers.odp_file import OdpFileProvider
from ..providers.ods_file import OdsFileProvider
from ..providers.odt_file import OdtFileProvider
from ..providers.pdf_file import GraphicPdfFileProvider, PdfFileProvider
from ..providers.png_file import GraphicPngFileProvider, PngFileProvider
from ..providers.pptx_file import PptxFileProvider
from ..providers.rtf_file import RtfFileProvider
from ..providers.svg_file import SvgFileProvider
from ..providers.tar_file import TarFileProvider
from ..providers.txt_file import TxtFileProvider
from ..providers.webp_file import GraphicWebpFileProvider, WebpFileProvider
from ..providers.xlsx_file import XlsxFileProvider
from ..providers.xml_file import XmlFileProvider
from ..providers.zip_file import ZipFileProvider

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "generate_completion_file",
    "generate_file",
    "get_method_kwargs",
    "is_optional_type",
    "PROVIDERS",
)

KWARGS_DROP = {
    "self",  # Drop as irrelevant
    "storage",  # Drop as non-supported arg
    "return",  # Drop as irrelevant
    # "mp3_generator_cls",  # Drop as non-supported arg
    # "mp3_generator_kwargs",  # Drop as non-supported arg
    # "pdf_generator_cls",  # Drop as non-supported arg
    # "pdf_generator_kwargs",  # Drop as non-supported arg
    "format_func",  # Drop as non-supported arg
    "raw",  # Drop `raw`, because we will be forcing raw=True for streaming
}
OVERRIDES = {
    "DocxFileProvider.docx_file": {
        "annotations": {
            "content": str,
        },
        "model_props": {
            "content": None,
        },
    },
    "GenericFileProvider.generic_file": {
        "annotations": {
            "content": str,
        },
        "model_props": {},
    },
    "Mp3FileProvider.mp3_file": {
        "annotations": {
            "mp3_generator_cls": str,
        },
        "model_props": {
            "mp3_generator_cls": (
                "faker_file.providers.mp3_file.generators"
                ".gtts_generator.GttsMp3Generator"
            ),
        },
    },
    "OdtFileProvider.odt_file": {
        "annotations": {
            "content": str,
        },
        "model_props": {
            "content": None,
        },
    },
    "PdfFileProvider.pdf_file": {
        "annotations": {
            "pdf_generator_cls": str,
        },
        "model_props": {
            "pdf_generator_cls": (
                "faker_file.providers.pdf_file.generators"
                ".pdfkit_generator.PdfkitPdfGenerator"
            ),
        },
    },
}
PROVIDERS = {
    BinFileProvider.bin_file.__name__: BinFileProvider,
    CsvFileProvider.csv_file.__name__: CsvFileProvider,
    DocxFileProvider.docx_file.__name__: DocxFileProvider,
    EmlFileProvider.eml_file.__name__: EmlFileProvider,
    EpubFileProvider.epub_file.__name__: EpubFileProvider,
    GenericFileProvider.generic_file.__name__: GenericFileProvider,
    GraphicIcoFileProvider.graphic_ico_file.__name__: GraphicIcoFileProvider,
    GraphicJpegFileProvider.graphic_jpeg_file.__name__: (
        GraphicJpegFileProvider
    ),
    GraphicPdfFileProvider.graphic_pdf_file.__name__: GraphicPdfFileProvider,
    GraphicPngFileProvider.graphic_png_file.__name__: GraphicPngFileProvider,
    GraphicWebpFileProvider.graphic_webp_file.__name__: (
        GraphicWebpFileProvider
    ),
    IcoFileProvider.ico_file.__name__: IcoFileProvider,
    JpegFileProvider.jpeg_file.__name__: JpegFileProvider,
    Mp3FileProvider.mp3_file.__name__: Mp3FileProvider,
    OdpFileProvider.odp_file.__name__: OdpFileProvider,
    OdsFileProvider.ods_file.__name__: OdsFileProvider,
    OdtFileProvider.odt_file.__name__: OdtFileProvider,
    PdfFileProvider.pdf_file.__name__: PdfFileProvider,
    PngFileProvider.png_file.__name__: PngFileProvider,
    PptxFileProvider.pptx_file.__name__: PptxFileProvider,
    RtfFileProvider.rtf_file.__name__: RtfFileProvider,
    SvgFileProvider.svg_file.__name__: SvgFileProvider,
    TarFileProvider.tar_file.__name__: TarFileProvider,
    TxtFileProvider.txt_file.__name__: TxtFileProvider,
    WebpFileProvider.webp_file.__name__: WebpFileProvider,
    XlsxFileProvider.xlsx_file.__name__: XlsxFileProvider,
    XmlFileProvider.xml_file.__name__: XmlFileProvider,
    ZipFileProvider.zip_file.__name__: ZipFileProvider,
}


def get_method_kwargs(
    cls: Type[FileMixin], method_name: str
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    method = getattr(cls, method_name)
    method_specs = inspect.getfullargspec(method)

    kwargs = deepcopy(method_specs.args[1:])  # Omit `self`
    defaults = deepcopy(method_specs.defaults if method_specs.defaults else [])
    model_props = dict(zip(kwargs, defaults))
    annotations = deepcopy(method_specs.annotations)

    # Override the type definition for mp3_generator_cls
    override = OVERRIDES.get(f"{cls.__name__}.{method_name}", None)
    if override:
        annotations.update(override["annotations"])
        model_props.update(override["model_props"])

    for kwarg_name in KWARGS_DROP:
        annotations.pop(kwarg_name, None)
        model_props.pop(kwarg_name, None)

    return model_props, annotations


def generate_file(method_name: str, **kwargs) -> StringValue:
    faker = Faker()
    cls = PROVIDERS[method_name]
    method = getattr(cls(faker), method_name)
    value = method(**kwargs)
    return value


def is_optional_type(t: Any) -> bool:
    if getattr(t, "__origin__", None) is typing.Union:
        return any(arg is type(None) for arg in t.__args__)  # noqa
    return False


def generate_completion_file():
    completion_script = f"""#!/bin/bash

_faker_file_completion() {{
    local cur prev providers commands
    cur="${{COMP_WORDS[COMP_CWORD]}}"
    prev="${{COMP_WORDS[COMP_CWORD - 1]}}"
    providers="{(' '.join(PROVIDERS.keys()))}"
    commands="generate-completion version"  # Add the commands here

    case $prev in"""

    for method_name, provider in PROVIDERS.items():
        method_kwargs, _ = get_method_kwargs(provider, method_name)
        completion_script += f"""
        {method_name})
            COMPREPLY=($(compgen -W "{(' '.join('--' + k for k in method_kwargs.keys()))} --nb_files" -- "$cur"))
            ;;
        """  # noqa

    # Add the case for commands
    completion_script += """
        generate-completion|version)
            COMPREPLY=()
            ;;
    """

    completion_script += """
        *)
            COMPREPLY=($(compgen -W "$providers $commands" -- "$cur"))
            ;;
    esac

    return 0
}

complete -F _faker_file_completion faker-file
"""

    user_home_dir = os.path.expanduser("~")
    file_path = os.path.join(user_home_dir, "faker_file_completion.sh")
    with open(file_path, "w") as f:
        f.write(completion_script)

    print(f"Generated bash completion file: {file_path}")



from abc import abstractmethod
from typing import Any, Dict, Optional, Union

from pathy import Pathy

from ..base import DEFAULT_REL_PATH
from .base import BaseStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = (
    "CloudStorage",
    "PathyFileSystemStorage",
)


DEFAULT_ROOT_PATH = "tmp"


class CloudStorage(BaseStorage):
    """Base cloud storage."""

    bucket_name: str
    bucket: Pathy
    credentials: Dict[str, str]
    schema: Optional[str] = None

    def __init__(
        self: "CloudStorage",
        bucket_name: str,
        root_path: Optional[str] = DEFAULT_ROOT_PATH,
        rel_path: Optional[str] = DEFAULT_REL_PATH,
        credentials: Optional[Dict[str, Any]] = None,
        *args,
        **kwargs,
    ):
        if self.schema is None:
            raise Exception("The `schema` property should the set!")
        self.bucket_name = bucket_name
        self.root_path = root_path or ""
        self.rel_path = rel_path or ""
        self.cache_dir = None
        credentials = credentials or {}

        if credentials:
            self.authenticate(**credentials)

        self.bucket = Pathy(f"{self.schema}://{self.bucket_name}")
        # If bucket does not exist, create
        if not self.bucket.exists():
            self.bucket.mkdir(exist_ok=True)

        super().__init__(*args, **kwargs)

    @abstractmethod
    def authenticate(self, *args, **kwargs):
        raise NotImplementedError("Method authenticate is not implemented!")

    def generate_filename(
        self: "CloudStorage",
        extension: str,
        prefix: Optional[str] = None,
        basename: Optional[str] = None,
    ) -> Pathy:
        """Generate filename."""
        if not extension:
            raise Exception("Extension shall be given!")

        if not basename:
            basename = self.generate_basename(prefix)

        return (
            self.bucket
            / self.root_path
            / self.rel_path
            / f"{basename}.{extension}"
        )

        # if basename:
        #     return (
        #         self.bucket
        #         / self.root_path
        #         / self.rel_path
        #         / f"{basename}.{extension}"
        #     )
        # else:
        #     with tempfile.NamedTemporaryFile(
        #         prefix=prefix,
        #         suffix=f".{extension}",
        #     ) as temp_file:
        #         return (
        #             self.bucket
        #             / self.root_path
        #             / self.rel_path
        #             / os.path.basename(temp_file.name)
        #         )

    def write_text(
        self: "CloudStorage",
        filename: Pathy,
        data: str,
        encoding: Optional[str] = None,
    ) -> int:
        """Write text."""
        file = self.bucket / self.root_path / self.rel_path / filename
        return file.write_text(data, encoding)

    def write_bytes(self: "CloudStorage", filename: Pathy, data: bytes) -> int:
        """Write bytes."""
        file = self.bucket / self.root_path / self.rel_path / filename
        return file.write_bytes(data)

    def exists(self: "CloudStorage", filename: Union[Pathy, str]) -> bool:
        """Check if file exists."""
        if isinstance(filename, str):
            filename = self.bucket / self.root_path / filename
        return filename.exists()

    def relpath(self: "CloudStorage", filename: Pathy) -> str:
        """Return relative path."""
        return str(filename.relative_to(self.bucket / self.root_path))

    def abspath(self: "CloudStorage", filename: Pathy) -> str:
        """Return relative path."""
        return filename.as_uri()

    def unlink(self: "CloudStorage", filename: Union[Pathy, str]) -> None:
        """Delete the file."""
        if isinstance(filename, str):
            filename = self.bucket / self.root_path / filename
        filename.unlink()


class PathyFileSystemStorage(CloudStorage):
    """Pathy FileSystem Storage.

    Usage example:

    .. code-block:: python

        from faker_file.storages.cloud import PathyFileSystemStorage

        fs_storage = PathyFileSystemStorage(bucket_name="artur-testing-1")
        file = fs_storage.generate_filename(prefix="zzz_", extension="docx")
        fs_storage.write_text(file, "Lorem ipsum")
        fs_storage.write_bytes(file, b"Lorem ipsum")
    """

    schema: str = "file"

    def authenticate(self: "PathyFileSystemStorage", **kwargs) -> None:
        """Authenticate. Does nothing."""



from io import BytesIO
from typing import Callable, Optional, Union, overload

from docx import Document
from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ..base import (
    DEFAULT_FORMAT_FUNC,
    BytesValue,
    DynamicTemplate,
    FileMixin,
    StringValue,
)
from ..constants import DEFAULT_TEXT_MAX_NB_CHARS
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("DocxFileProvider",)


class DocxFileProvider(BaseProvider, FileMixin):
    """DOCX file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.docx_file import DocxFileProvider

        FAKER = Faker()
        FAKER.add_provider(DocxFileProvider)

        file = FAKER.docx_file()

    Usage example with options:

    .. code-block:: python

        file = FAKER.docx_file(
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.docx_file(
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
            prefix="zzz",
            max_nb_chars=100_000,
            wrap_chars_after=80,
        )

    Usage example with content modifiers:

    .. code-block:: python

        from faker_file.base import DynamicTemplate
        from faker_file.contrib.docx_file import (
            add_h1_heading,
            add_h2_heading,
            add_h3_heading,
            add_h4_heading,
            add_h5_heading,
            add_h6_heading,
            add_page_break,
            add_paragraph,
            add_picture,
            add_table,
            add_title_heading,
        )

        file = FAKER.docx_file(
            content=DynamicTemplate(
                [
                    (add_title_heading, {}),
                    (add_paragraph, {}),
                    (add_h1_heading, {}),
                    (add_h2_heading, {}),
                    (add_h3_heading, {}),
                    (add_h4_heading, {}),
                    (add_h5_heading, {}),
                    (add_h6_heading, {}),
                    (add_paragraph, {}),
                    (add_picture, {}),
                    (add_page_break, {}),
                    (add_h6_heading, {}),
                    (add_table, {}),
                    (add_paragraph, {}),
                ]
            )
        )
    """

    extension: str = "docx"

    @overload
    def docx_file(
        self: "DocxFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[Union[str, DynamicTemplate]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def docx_file(
        self: "DocxFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[Union[str, DynamicTemplate]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def docx_file(
        self: "DocxFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        max_nb_chars: int = DEFAULT_TEXT_MAX_NB_CHARS,
        wrap_chars_after: Optional[int] = None,
        content: Optional[Union[str, DynamicTemplate]] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate a DOCX file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param max_nb_chars: Max number of chars for the content.
        :param wrap_chars_after: If given, the output string would be separated
             by line breaks after the given position.
        :param content: File content. Might contain dynamic elements (still
            being a string), which are then replaced by correspondent
            fixtures. Can alternatively be a `DynamicTemplate` - list of content
            modifiers (callables to call after the document instance has been
            created). Each callable should accept the following
            arguments: provider, document, data, counter and **kwargs.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        if isinstance(content, DynamicTemplate):
            _content = ""
        else:
            _content = self._generate_text_content(
                max_nb_chars=max_nb_chars,
                wrap_chars_after=wrap_chars_after,
                content=content,
                format_func=format_func,
            )
        data = {"content": _content, "filename": filename, "storage": storage}

        with BytesIO() as _fake_file:
            document = Document()
            if _content:
                document.add_paragraph(_content)
            elif isinstance(content, DynamicTemplate):
                for counter, (ct_modifier, ct_modifier_kwargs) in enumerate(
                    content.content_modifiers
                ):
                    ct_modifier(
                        self,
                        document,
                        data,
                        counter,
                        **ct_modifier_kwargs,
                    )

            document.save(_fake_file)

            if raw:
                raw_content = BytesValue(_fake_file.getvalue())
                raw_content.data = data
                return raw_content

            storage.write_bytes(filename, _fake_file.getvalue())

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from faker import Faker
from faker_file.providers.bin_file import BinFileProvider

FAKER = Faker()
FAKER.add_provider(BinFileProvider)

bin_file = FAKER.bin_file(length=1024**2)  # 1 Mb
bin_file = FAKER.bin_file(length=3 * 1024**2)  # 3 Mb
bin_file = FAKER.bin_file(length=10 * 1024**2)  # 10 Mb

bin_file = FAKER.bin_file(length=1024)  # 1 Kb
bin_file = FAKER.bin_file(length=3 * 1024)  # 3 Kb
bin_file = FAKER.bin_file(length=10 * 1024)  # 10 Kb



import logging
import os
import socket
import tempfile
import threading
import time
import unittest
from functools import partial
from typing import Any, Dict, Type, Union

from faker import Faker
from parametrize import parametrize

from ..providers.txt_file import TxtFileProvider
from ..registry import FILE_REGISTRY
from ..storages.sftp_storage import SFTPStorage
from .sftp_server import SFTPServerManager, start_server
from .utils import AutoFreePortInt

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("TestSFTPStorageTestCase",)

SFTP_USER = os.environ.get("SFTP_USER", "foo")
SFTP_PASS = os.environ.get("SFTP_PASS", "pass")
SFTP_HOST = os.environ.get("SFTP_HOST", "127.0.0.1")
SFTP_PORT = int(os.environ.get("SFTP_PORT", AutoFreePortInt(host=SFTP_HOST)))
SFTP_ROOT_PATH = os.environ.get("SFTP_ROOT_PATH", "/upload")

LOGGER = logging.getLogger(__name__)

FAKER = Faker()
FAKER.add_provider(TxtFileProvider)


class TestSFTPStorageTestCase(unittest.TestCase):
    """Test SFTP storage."""

    server_manager: SFTPServerManager
    server_thread: threading.Thread
    sftp_host: str = SFTP_HOST
    sftp_port: int = int(AutoFreePortInt(host=SFTP_HOST))
    sftp_user: str = SFTP_USER
    sftp_pass: str = SFTP_PASS
    sftp_root_path: str = SFTP_ROOT_PATH
    # Maximum number of retries to check if the port is free
    max_port_retry_limit: int = 10

    @staticmethod
    def is_port_in_use(host: str, port: int) -> bool:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex((host, port)) == 0

    @classmethod
    def free_port(cls: "TestSFTPStorageTestCase") -> None:
        """Check if the port is in use and wait until it is free."""
        retry_count = 0
        while cls.is_port_in_use(cls.sftp_host, cls.sftp_port):
            if retry_count >= cls.max_port_retry_limit:
                LOGGER.warning(
                    f"Maximum retries reached. Port {cls.sftp_port} on "
                    f"host {cls.sftp_host} may not be free soon."
                )
                retry_count = 0  # Reset the retry count
                # Assign a new port
                cls.sftp_port = int(AutoFreePortInt(host=SFTP_HOST))
                LOGGER.warning(
                    f"Assigned a new port {cls.sftp_port} on "
                    f"host {cls.sftp_host}. Restarting attempts..."
                )
                continue

            LOGGER.info(
                f"Port {cls.sftp_port} in use on host {cls.sftp_host}, "
                f"waiting..."
            )
            time.sleep(1)
            retry_count += 1

    def tearDown(self) -> None:
        super().tearDown()
        # FILE_REGISTRY.clean_up()

    @classmethod
    def setUpClass(cls):
        # Free port
        cls.free_port()

        os.makedirs(
            os.path.join(tempfile.gettempdir(), "upload", "sub"),
            exist_ok=True,
        )

        # Create a partial function with custom arguments
        start_server_partial = partial(
            start_server, host=cls.sftp_host, port=cls.sftp_port
        )

        # Start the server in a separate thread
        cls.server_thread = threading.Thread(target=start_server_partial)
        # Daemonize the thread, so it exits when the main thread exits
        cls.server_thread.daemon = True
        cls.server_thread.start()
        time.sleep(2)

    @parametrize(
        "storage_cls, kwargs, prefix, basename, extension",
        [
            # SFTPStorage
            (
                SFTPStorage,
                {
                    "host": sftp_host,
                    "port": sftp_port,
                    "username": sftp_user,
                    "password": sftp_pass,
                    "root_path": sftp_root_path,
                },
                "zzz",
                None,
                "docx",
            ),
            (
                SFTPStorage,
                {
                    "host": sftp_host,
                    "port": sftp_port,
                    "username": sftp_user,
                    "password": sftp_pass,
                    "root_path": sftp_root_path,
                },
                None,
                "my_zzz_filename",
                "docx",
            ),
        ],
    )
    def test_storage(
        self: "TestSFTPStorageTestCase",
        storage_cls: Type[SFTPStorage],
        kwargs: Dict[str, Any],
        prefix: Union[str, None],
        basename: Union[str, None],
        extension: str,
    ) -> None:
        """Test storage."""
        storage = storage_cls(**kwargs)
        # Text file
        file_text = storage.generate_filename(
            basename=basename, prefix=prefix, extension=extension
        )
        # Write to the text file
        text_result = storage.write_text(file_text, "Lorem ipsum")
        # Check if file exists
        self.assertTrue(storage.exists(file_text))
        # Assert correct return value
        self.assertIsInstance(text_result, int)

        # Bytes
        file_bytes = storage.generate_filename(
            basename=basename, prefix=prefix, extension=extension
        )
        # Write to bytes file
        bytes_result = storage.write_bytes(file_bytes, b"Lorem ipsum")
        # Check if file exists
        self.assertTrue(storage.exists(file_bytes))
        # Assert correct return value
        self.assertIsInstance(bytes_result, int)

        # Clean up
        storage.unlink(file_text)
        storage.unlink(file_bytes)

    @parametrize(
        "storage_cls, kwargs, prefix, extension",
        [
            # FileSystemStorage
            (
                SFTPStorage,
                {
                    "host": sftp_host,
                    "port": sftp_port,
                    "username": sftp_user,
                    "password": sftp_pass,
                    "root_path": sftp_root_path,
                },
                "zzz",
                "",
            ),
        ],
    )
    def test_storage_generate_filename_exceptions(
        self: "TestSFTPStorageTestCase",
        storage_cls: Type[SFTPStorage],
        kwargs: Dict[str, Any],
        prefix: str,
        extension: str,
    ) -> None:
        """Test storage `generate_filename` exceptions."""
        storage = storage_cls(**kwargs)

        with self.assertRaises(Exception):
            # Generate filename
            storage.generate_filename(prefix=prefix, extension=extension)

        with self.assertRaises(Exception):
            # Generate filename
            storage.generate_filename(basename=prefix, extension=extension)

    def test_file_system_storage_abspath(
        self: "TestSFTPStorageTestCase",
    ) -> None:
        """Test `FileSystemStorage` `abspath`."""

        storage = SFTPStorage(
            host=self.sftp_host,
            port=self.sftp_port,
            username=self.sftp_user,
            password=self.sftp_pass,
            root_path=self.sftp_root_path,
        )
        filename = storage.generate_filename(prefix="", extension="tmp")
        self.assertTrue(
            storage.abspath(filename).startswith(self.sftp_root_path)
        )

    def test_integration(self: "TestSFTPStorageTestCase") -> None:
        storage = SFTPStorage(
            host=self.sftp_host,
            port=self.sftp_port,
            username=self.sftp_user,
            password=self.sftp_pass,
            root_path=self.sftp_root_path,
        )
        file = FAKER.txt_file(storage=storage)
        self.assertTrue(storage.exists(file.data["filename"]))
        self.assertTrue(
            storage.abspath(file.data["filename"]).startswith(
                self.sftp_root_path
            )
        )

        FILE_REGISTRY.clean_up()  # Clean up storage files

    def test_integration_sub_dir(self: "TestSFTPStorageTestCase") -> None:
        storage = SFTPStorage(
            host=self.sftp_host,
            port=self.sftp_port,
            username=self.sftp_user,
            password=self.sftp_pass,
            root_path=self.sftp_root_path,
            rel_path="sub",
        )
        file = FAKER.txt_file(storage=storage)
        self.assertTrue(storage.exists(file.data["filename"]))
        self.assertTrue(
            storage.abspath(file.data["filename"]).startswith(
                os.path.join(self.sftp_root_path, "sub")
            )
        )

        FILE_REGISTRY.clean_up()  # Clean up storage files

    @parametrize(
        "kwargs",
        [
            # Wrong username, key
            (
                {
                    "host": sftp_host,
                    "port": sftp_port,
                    "username": "wrong",
                    "key": "wrong",
                },
            ),
            # Wrong username, password
            (
                {
                    "host": sftp_host,
                    "port": sftp_port,
                    "username": "wrong",
                    "password": "wrong",
                },
            ),
        ],
    )
    def test_storage_initialization_exceptions(
        self: "TestSFTPStorageTestCase",
        kwargs: Dict[str, Any],
    ) -> None:
        with self.assertRaises(Exception):
            # Initialize the storage
            SFTPStorage(**kwargs)

    def test_storage_write_text_exceptions(self) -> None:
        storage = SFTPStorage(
            host=self.sftp_host,
            port=self.sftp_port,
            username=self.sftp_user,
            password=self.sftp_pass,
            root_path=self.sftp_root_path,
        )
        val = storage.write_text(
            filename=FAKER.file_name(),
            data=1,  # type: ignore
        )
        self.assertEqual(val, -1)

    def test_storage_write_bytes_exceptions(self) -> None:
        storage = SFTPStorage(
            host=self.sftp_host,
            port=self.sftp_port,
            username=self.sftp_user,
            password=self.sftp_pass,
            root_path=self.sftp_root_path,
        )
        val = storage.write_bytes(
            filename=FAKER.file_name(),
            data=1,  # type: ignore
        )
        self.assertEqual(val, -1)

    def test_storage_exists_exceptions(self) -> None:
        storage = SFTPStorage(
            host=self.sftp_host,
            port=self.sftp_port,
            username=self.sftp_user,
            password=self.sftp_pass,
            root_path=self.sftp_root_path,
        )
        val = storage.exists(
            filename=FAKER.file_name(),
        )
        self.assertFalse(val)



# Imports and initialization
from faker import Faker
from faker_file.providers.augment_file_from_dir import (
    AugmentFileFromDirProvider,
)
from faker_file.providers.bin_file import BinFileProvider
from faker_file.providers.bmp_file import BmpFileProvider
from faker_file.providers.csv_file import CsvFileProvider
from faker_file.providers.docx_file import DocxFileProvider
from faker_file.providers.eml_file import EmlFileProvider
from faker_file.providers.epub_file import EpubFileProvider
from faker_file.providers.gif_file import GifFileProvider
from faker_file.providers.ico_file import (
    GraphicIcoFileProvider,
    IcoFileProvider,
)
from faker_file.providers.jpeg_file import (
    GraphicJpegFileProvider,
    JpegFileProvider,
)
from faker_file.providers.mp3_file import Mp3FileProvider
from faker_file.providers.odp_file import OdpFileProvider
from faker_file.providers.ods_file import OdsFileProvider
from faker_file.providers.odt_file import OdtFileProvider
from faker_file.providers.pdf_file import (
    GraphicPdfFileProvider,
    PdfFileProvider,
)
from faker_file.providers.png_file import (
    GraphicPngFileProvider,
    PngFileProvider,
)
from faker_file.providers.pptx_file import PptxFileProvider
from faker_file.providers.random_file_from_dir import RandomFileFromDirProvider
from faker_file.providers.rtf_file import RtfFileProvider
from faker_file.providers.svg_file import SvgFileProvider
from faker_file.providers.tar_file import TarFileProvider
from faker_file.providers.tiff_file import TiffFileProvider
from faker_file.providers.txt_file import TxtFileProvider
from faker_file.providers.webp_file import (
    GraphicWebpFileProvider,
    WebpFileProvider,
)
from faker_file.providers.xlsx_file import XlsxFileProvider
from faker_file.providers.zip_file import ZipFileProvider

FAKER = Faker()
FAKER.add_provider(AugmentFileFromDirProvider)
FAKER.add_provider(BinFileProvider)
FAKER.add_provider(BmpFileProvider)
FAKER.add_provider(CsvFileProvider)
FAKER.add_provider(DocxFileProvider)
FAKER.add_provider(EmlFileProvider)
FAKER.add_provider(EpubFileProvider)
FAKER.add_provider(GifFileProvider)
FAKER.add_provider(GraphicIcoFileProvider)
FAKER.add_provider(GraphicJpegFileProvider)
FAKER.add_provider(GraphicPdfFileProvider)
FAKER.add_provider(GraphicPngFileProvider)
FAKER.add_provider(GraphicWebpFileProvider)
FAKER.add_provider(IcoFileProvider)
FAKER.add_provider(JpegFileProvider)
FAKER.add_provider(Mp3FileProvider)
FAKER.add_provider(OdpFileProvider)
FAKER.add_provider(OdsFileProvider)
FAKER.add_provider(OdtFileProvider)
FAKER.add_provider(PdfFileProvider)
FAKER.add_provider(PngFileProvider)
FAKER.add_provider(PptxFileProvider)
FAKER.add_provider(RandomFileFromDirProvider)
FAKER.add_provider(RtfFileProvider)
FAKER.add_provider(SvgFileProvider)
FAKER.add_provider(TarFileProvider)
FAKER.add_provider(TiffFileProvider)
FAKER.add_provider(TxtFileProvider)
FAKER.add_provider(WebpFileProvider)
FAKER.add_provider(XlsxFileProvider)
FAKER.add_provider(ZipFileProvider)

# Create files to test `augment_file_from_dir` with
FAKER.docx_file()
FAKER.eml_file()
FAKER.odt_file()
FAKER.txt_file()

# Usage examples
# augmented_file = FAKER.augment_file_from_dir(
#     source_dir_path="/tmp/tmp/",
# )
bin_file = FAKER.bin_file()
bmp_file = FAKER.bmp_file()
csv_file = FAKER.csv_file()
docx_file = FAKER.docx_file()
eml_file = FAKER.eml_file()
epub_file = FAKER.epub_file()
gif_file = FAKER.gif_file()
graphic_ico_file = FAKER.graphic_ico_file()
graphic_jpeg_file = FAKER.graphic_jpeg_file()
graphic_pdf_file = FAKER.graphic_pdf_file()
graphic_png_file = FAKER.graphic_png_file()
graphic_webp_file = FAKER.graphic_webp_file()
ico_file = FAKER.ico_file()
jpeg_file = FAKER.jpeg_file()
mp3_file = FAKER.mp3_file()
odp_file = FAKER.odp_file()
ods_file = FAKER.ods_file()
odt_file = FAKER.odt_file()
pdf_file = FAKER.pdf_file()
png_file = FAKER.png_file()
pptx_file = FAKER.pptx_file()
random_file = FAKER.random_file_from_dir(
    source_dir_path="/tmp/tmp/",
)
rtf_file = FAKER.rtf_file()
svg_file = FAKER.svg_file()
tar_file = FAKER.tar_file()
tiff_file = FAKER.tiff_file()
txt_file = FAKER.txt_file()
# webp_file = FAKER.webp_file()
xlsx_file = FAKER.xlsx_file()
zip_file = FAKER.zip_file()

# Usage examples bytes back
# augmented_raw = FAKER.augment_file_from_dir(
#     source_dir_path="/tmp/tmp/",
#     raw=True,
# )
bin_raw = FAKER.bin_file(raw=True)
bmp_raw = FAKER.bmp_file(raw=True)
csv_raw = FAKER.csv_file(raw=True)
docx_raw = FAKER.docx_file(raw=True)
eml_raw = FAKER.eml_file(raw=True)
epub_raw = FAKER.epub_file(raw=True)
gif_raw = FAKER.gif_file(raw=True)
ico_raw = FAKER.ico_file(raw=True)
jpeg_raw = FAKER.jpeg_file(raw=True)
mp3_raw = FAKER.mp3_file(raw=True)
odp_raw = FAKER.odp_file(raw=True)
ods_raw = FAKER.ods_file(raw=True)
odt_raw = FAKER.odt_file(raw=True)
pdf_raw = FAKER.pdf_file(raw=True)
png_raw = FAKER.png_file(raw=True)
pptx_raw = FAKER.pptx_file(raw=True)
random_raw = FAKER.random_file_from_dir(
    source_dir_path="/tmp/tmp/",
    raw=True,
)
rtf_raw = FAKER.rtf_file(raw=True)
svg_raw = FAKER.svg_file(raw=True)
tar_raw = FAKER.tar_file(raw=True)
tiff_raw = FAKER.tiff_file(raw=True)
txt_raw = FAKER.txt_file(raw=True)
# webp_raw = FAKER.webp_file(raw=True)
xlsx_raw = FAKER.xlsx_file(raw=True)
zip_raw = FAKER.zip_file(raw=True)



import logging
import os
from pathlib import Path
from random import choice
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    overload,
)

from faker.providers import BaseProvider

from ..base import BytesValue, FileMixin, StringValue
from ..helpers import random_pop
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage
from .image.augment import augment_image_file

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("AugmentRandomImageFromDirProvider",)

LOGGER = logging.getLogger(__name__)

EXTENSIONS = {
    "bmp",
    "gif",
    # "ico",  # Not supported yet
    "jpeg",
    "jpg",
    "png",
    # "svg",  # Not supported yet
    "tiff",
    "webp",
}


class AugmentRandomImageFromDirProvider(BaseProvider, FileMixin):
    """Augment image from given directory provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.augment_random_image_from_dir import (
            AugmentRandomImageFromDirProvider
        )

        FAKER = Faker()
        FAKER.add_provider(AugmentRandomImageFromDirProvider)

        file = FAKER.augment_random_image_from_dir(
            source_dir_path="/tmp/tmp/"
        )

    Usage example with options:

    .. code-block:: python

        file = FAKER.augment_random_image_from_dir(
            source_dir_path="/tmp/tmp/",
            prefix="zzz",
            extensions={"jpeg", "png"},
        )
    """

    extension: str = ""

    @overload
    def augment_random_image_from_dir(
        self: "AugmentRandomImageFromDirProvider",
        source_dir_path: str,
        extensions: Optional[Iterable[str]] = None,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        augmentations: Optional[List[Tuple[Callable, Dict[str, Any]]]] = None,
        num_steps: Optional[int] = None,
        pop_func: Callable = random_pop,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def augment_random_image_from_dir(
        self: "AugmentRandomImageFromDirProvider",
        source_dir_path: str,
        extensions: Optional[Iterable[str]] = None,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        augmentations: Optional[List[Tuple[Callable, Dict[str, Any]]]] = None,
        num_steps: Optional[int] = None,
        pop_func: Callable = random_pop,
        **kwargs,
    ) -> StringValue: ...

    def augment_random_image_from_dir(
        self: "AugmentRandomImageFromDirProvider",
        source_dir_path: str,
        extensions: Optional[Iterable[str]] = None,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        augmentations: Optional[List[Tuple[Callable, Dict[str, Any]]]] = None,
        num_steps: Optional[int] = None,
        pop_func: Callable = random_pop,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Augment a random image from given directory.

        :param source_dir_path: Source files directory.
        :param extensions: Allowed extensions.
        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param augmentations: List of tuples of callable augmentation
            functions and their respective keyword arguments. If not
            provided, the default augmentation functions will be used.
        :param num_steps: Number of augmentation steps (functions) to be
            applied. If not specified, the length of the `augmentations` list
            will be used.
        :param pop_func: Callable to pop items from `augmentations` list. By
            default, the `random_pop` is used, which pops items in random
            order. If you want the order of augmentations to be constant and
            as given, replace it with `list.pop` (`pop_func=list.pop`).
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """
        # Generic
        if storage is None:
            storage = FileSystemStorage()

        if extensions is None:
            extensions = EXTENSIONS

        # Specific
        source_file_choices = [
            os.path.join(source_dir_path, _f)
            for _f in os.listdir(source_dir_path)
            if (
                os.path.isfile(os.path.join(source_dir_path, _f))
                and os.path.splitext(_f)[1][1:] in extensions
            )
        ]
        source_file_path = choice(source_file_choices)
        source_file = Path(source_file_path)

        # Generic
        filename = storage.generate_filename(
            extension=source_file.suffix[1:],
            prefix=prefix,
            basename=basename,
        )
        data = {"filename": filename, "storage": storage}

        image_bytes = augment_image_file(
            image_path=source_file_path,
            augmentations=augmentations,
            num_steps=num_steps,
            pop_func=pop_func,
        )

        if raw:
            raw_content = BytesValue(image_bytes)
            raw_content.data = data
            return raw_content

        storage.write_bytes(filename, image_bytes)

        # Generic
        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from faker import Faker
from faker_file.providers.helpers.inner import (
    create_inner_docx_file,
    create_inner_zip_file,
)
from faker_file.providers.zip_file import ZipFileProvider

FAKER = Faker()
FAKER.add_provider(ZipFileProvider)

zip_file = FAKER.zip_file(
    prefix="nested_level_0_",
    options={
        "create_inner_file_func": create_inner_zip_file,
        "create_inner_file_args": {
            "prefix": "nested_level_1_",
            "options": {
                "create_inner_file_func": create_inner_zip_file,
                "create_inner_file_args": {
                    "prefix": "nested_level_2_",
                    "options": {
                        "create_inner_file_func": create_inner_docx_file,
                    },
                },
            },
        },
    },
)



from pathlib import Path
from typing import Union

import tika
from tika import parser

from ...base.text_extractor import BaseTextExtractor

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("TikaTextExtractor",)


class TikaTextExtractor(BaseTextExtractor):
    """Text extractor based on `Apache Tika` and `tika-python`.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.augment_file_from_dir import (
            AugmentFileFromDirProvider,
        )
        from faker_file.providers.augment_file_from_dir.extractors import (
            tika_extractor,
        )

        FAKER = Faker()
        FAKER.add_provider(AugmentFileFromDirProvider)

        file = FAKER.augment_file_from_dir(
            text_extractor_cls=tika_extractor.TikaTextExtractor
        )
    """

    def handle_kwargs(self: "TikaTextExtractor", **kwargs) -> None:
        """Handle kwargs."""

    def extract(
        self: "TikaTextExtractor",
        source_file: Union[Path, str],
    ) -> str:
        """Extract text."""
        tika.initVM()
        parsed = parser.from_file(str(source_file))
        return parsed["content"]



from faker import Faker
from faker_file.providers.image.imgkit_generator import ImgkitImageGenerator
from faker_file.providers.png_file import PngFileProvider

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(PngFileProvider)  # Register PngFileProvider

# Generate an image file, wrapping each line after 80 characters
png_file = FAKER.png_file(
    image_generator_cls=ImgkitImageGenerator, wrap_chars_after=80
)



import io
import logging
from typing import Any, Dict, Union

from faker import Faker
from faker.generator import Generator
from faker.providers.python import Provider
from pdf2image import convert_from_bytes
from PIL import Image
from weasyprint import HTML

from ...base import DynamicTemplate, StringList
from ...constants import DEFAULT_FILE_ENCODING
from ..base.image_generator import BaseImageGenerator

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("WeasyPrintImageGenerator",)

LOGGER = logging.getLogger(__name__)


class WeasyPrintImageGenerator(BaseImageGenerator):
    """WeasyPrint and Pdf2Image ImageGenerator image generator.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.png_file import PngFileProvider
        from faker_file.providers.image.weasyprint_generator import (
            WeasyPrintImageGenerator
        )

        FAKER = Faker()
        FAKER.add_provider(PngFileProvider)

        file = FAKER.png_file(
            img_generator_cls=WeasyPrintImageGenerator
        )

    With dynamic content:

    .. code-block:: python

        from faker import Faker
        from faker_file.base import DynamicTemplate
        from faker_file.contrib.image.weasyprint_snippets import *
        from faker_file.providers.image.weasyprint_generator import (
            WeasyPrintImageGenerator
        )
        from faker_file.providers.png_file import PngFileProvider

        FAKER = Faker()
        FAKER.add_provider(PngFileProvider)

        file = FAKER.png_file(
            image_generator_cls=WeasyPrintImageGenerator,
            content=DynamicTemplate(
                [
                    (add_h1_heading, {}),
                    (add_paragraph, {"max_nb_chars": 500}),
                    (add_paragraph, {"max_nb_chars": 500}),
                    (add_paragraph, {"max_nb_chars": 500}),
                    (add_paragraph, {"max_nb_chars": 500}),
                ]
            )
        )

        file = FAKER.png_file(
            image_generator_cls=WeasyPrintImageGenerator,
            content=DynamicTemplate(
                [
                    (add_h1_heading, {}),
                    (add_paragraph, {}),
                    (add_picture, {}),
                    (add_paragraph, {}),
                    (add_picture, {}),
                    (add_paragraph, {}),
                    (add_picture, {}),
                    (add_paragraph, {}),
                ]
            )
        )

        file = FAKER.png_file(
            image_generator_cls=WeasyPrintImageGenerator,
            content=DynamicTemplate(
                [
                    (add_h1_heading, {}),
                    (add_picture, {}),
                    (add_paragraph, {"max_nb_chars": 500}),
                    (add_picture, {}),
                    (add_paragraph, {"max_nb_chars": 500}),
                    (add_picture, {}),
                    (add_paragraph, {"max_nb_chars": 500}),
                    (add_picture, {}),
                    (add_paragraph, {"max_nb_chars": 500}),
                ]
            )
        )

        file = FAKER.png_file(
            image_generator_cls=WeasyPrintImageGenerator,
            content=DynamicTemplate(
                [
                    (add_h1_heading, {}),
                    (add_picture, {}),
                    (add_paragraph, {"max_nb_chars": 500}),
                    (add_table, {"rows": 5, "cols": 4}),
                ]
            )
        )

        file = FAKER.png_file(
            image_generator_cls=WeasyPrintImageGenerator,
            content=DynamicTemplate(
                [
                    (add_h1_heading, {"margin": (2, 2)}),
                    (add_picture, {"margin": (2, 2)}),
                    (add_paragraph, {"max_nb_chars": 500, "margin": (2, 2)}),
                    (add_picture, {"margin": (2, 2)}),
                    (add_paragraph, {"max_nb_chars": 500, "margin": (2, 2)}),
                    (add_picture, {"margin": (2, 2)}),
                    (add_paragraph, {"max_nb_chars": 500, "margin": (2, 2)}),
                    (add_picture, {"margin": (2, 2)}),
                    (add_paragraph, {"max_nb_chars": 500, "margin": (2, 2)}),
                ]
            )
        )
    """

    encoding: str = DEFAULT_FILE_ENCODING
    page_width: int = 794  # A4 size at 96 DPI
    page_height: int = 1123  # A4 size at 96 DPI
    wrapper_tag: str = "div"

    def __init__(self: "WeasyPrintImageGenerator", **kwargs):
        super().__init__(**kwargs)
        self.pages = []
        self.image_mode = "RGB"

    def handle_kwargs(self: "WeasyPrintImageGenerator", **kwargs) -> None:
        """Handle kwargs."""
        if "page_width" in kwargs:
            self.page_width = kwargs["page_width"]
        if "page_height" in kwargs:
            self.page_height = kwargs["page_height"]
        if "image_mode" in kwargs:
            self.image_mode = kwargs["image_mode"]
        if "wrapper_tag" in kwargs:
            self.wrapper_tag = kwargs["wrapper_tag"]

    def create_image_instance(
        self: "WeasyPrintImageGenerator",
        width: Union[int, None] = None,
        height: Union[int, None] = None,
    ) -> Image:
        return Image.new(
            self.image_mode,
            (width or self.page_width, height or self.page_height),
            # (width, height),
            (255, 255, 255),
        )

    def wrap(self: "WeasyPrintImageGenerator", content: str) -> str:
        return f"<{self.wrapper_tag}>" f"{content}" f"</{self.wrapper_tag}>"

    def generate(
        self: "WeasyPrintImageGenerator",
        content: str,
        data: Dict[str, Any],
        provider: Union[Faker, Generator, Provider],
        **kwargs,
    ) -> bytes:
        """Generate image."""
        if isinstance(content, DynamicTemplate):
            _content = StringList()
            for counter, (ct_modifier, ct_modifier_kwargs) in enumerate(
                content.content_modifiers
            ):
                ct_modifier(
                    provider,
                    self,
                    _content,
                    data,
                    counter,
                    **ct_modifier_kwargs,
                )
        else:
            # Convert the HTML to a PDF and store in memory
            pdf_bytes = HTML(
                string=f"<{self.wrapper_tag}>{content}</{self.wrapper_tag}>"
            ).write_pdf()

            # Convert the in-memory PDF to images
            self.pages = convert_from_bytes(pdf_bytes)

        # If the PDF has multiple pages, stitch them together
        if len(self.pages) > 1:
            # Get dimensions of combined image
            width, height = self.pages[0].size[0], sum(
                [img.size[1] for img in self.pages]
            )

            # Create a new image with white background
            result = self.create_image_instance(width, height)

            # Iterate over images and paste them onto the new image
            y_offset = 0
            for img in self.pages:
                result.paste(img, (0, y_offset))
                y_offset += img.size[1]
        else:
            # If there's only one page, use it directly
            result = self.pages[0]

        # Save the result to a BytesIO object
        output = io.BytesIO()
        result.save(output, format=provider.image_format)
        image_bytes = output.getvalue()  # Image, stored in memory

        return image_bytes



# Required imports
from faker import Faker
from faker_file.providers.pdf_file import PdfFileProvider
from faker_file.providers.pdf_file.generators.reportlab_generator import (
    ReportlabPdfGenerator,
)

FAKER = Faker()  # Initialize Faker
FAKER.add_provider(PdfFileProvider)  # Register provider

# Generate PDF file using `reportlab`
pdf_file = FAKER.pdf_file(pdf_generator_cls=ReportlabPdfGenerator)



import xml.etree.ElementTree as ET
from typing import Callable, Dict, Optional, Union, overload

from faker import Faker
from faker.generator import Generator
from faker.providers import BaseProvider
from faker.providers.python import Provider

from ..base import DEFAULT_FORMAT_FUNC, BytesValue, FileMixin, StringValue
from ..constants import DEFAULT_XML_DATA_COLUMNS
from ..registry import FILE_REGISTRY
from ..storages.base import BaseStorage
from ..storages.filesystem import FileSystemStorage

__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2022-2023 Artur Barseghyan"
__license__ = "MIT"
__all__ = ("XmlFileProvider",)

FAKER = Faker()


class XmlFileProvider(BaseProvider, FileMixin):
    """XML file provider.

    Usage example:

    .. code-block:: python

        from faker import Faker
        from faker_file.providers.xml_file import XmlFileProvider

        FAKER = Faker()
        FAKER.add_provider(XmlFileProvider)

        file = FAKER.xml_file()

    Usage example with options:

    .. code-block:: python

        file = FAKER.xml_file(
            prefix="zzz",
            num_rows=100,
            data_columns={
                "name": "{{name}}",
                "sentence": "{{sentence}}",
                "address": "{{address}}",
            },
        )

    Usage example with `FileSystemStorage` storage (for `Django`):

    .. code-block:: python

        from django.conf import settings
        from faker_file.storages.filesystem import FileSystemStorage

        file = FAKER.xml_file(
            storage=FileSystemStorage(
                root_path=settings.MEDIA_ROOT,
                rel_path="tmp",
            ),
            prefix="zzz",
            num_rows=100,
        )

    Usage example with template:

    .. code-block:: python

        XML_TEMPLATE = '''
        <books>
            <book>
                <name>{{sentence}}</name>
                <description>{{paragraph}}</description>
                <isbn>{{isbn13}}</isbn>
            </book>
            <book>
                <name>{{sentence}}</name>
                <description>{{paragraph}}</description>
                <isbn>{{isbn13}}</isbn>
            </book>
            <book>
                <name>{{sentence}}</name>
                <description>{{paragraph}}</description>
                <isbn>{{isbn13}}</isbn>
            </book>
        </books>
        '''

        file = FAKER.xml_file(content=XML_TEMPLATE)
    """

    extension: str = "xml"

    def _generate_xml_element(
        self,
        element_name: str,
        content_template: str,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
    ) -> ET.Element:
        element = ET.Element(element_name)
        element.text = format_func(self.generator, content_template)
        return element

    @overload
    def xml_file(
        self: "XmlFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        root_element: str = "root",
        row_element: str = "row",
        data_columns: Optional[Dict[str, str]] = None,
        num_rows: int = 10,
        content: Optional[str] = None,
        encoding: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = True,
        **kwargs,
    ) -> BytesValue: ...

    @overload
    def xml_file(
        self: "XmlFileProvider",
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        root_element: str = "root",
        row_element: str = "row",
        data_columns: Optional[Dict[str, str]] = None,
        num_rows: int = 10,
        content: Optional[str] = None,
        encoding: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        **kwargs,
    ) -> StringValue: ...

    def xml_file(
        self,
        storage: Optional[BaseStorage] = None,
        basename: Optional[str] = None,
        prefix: Optional[str] = None,
        root_element: str = "root",
        row_element: str = "row",
        data_columns: Optional[Dict[str, str]] = None,
        num_rows: int = 10,
        content: Optional[str] = None,
        encoding: Optional[str] = None,
        format_func: Callable[
            [Union[Faker, Generator, Provider], str], str
        ] = DEFAULT_FORMAT_FUNC,
        raw: bool = False,
        **kwargs,
    ) -> Union[BytesValue, StringValue]:
        """Generate an XML file with random text.

        :param storage: Storage. Defaults to `FileSystemStorage`.
        :param basename: File basename (without extension).
        :param prefix: File name prefix.
        :param root_element: Root XML element.
        :param row_element: Row XML element.
        :param data_columns: The ``data_columns`` argument expects a list or a
            tuple of string tokens, and these string tokens will be passed to
            :meth:`pystr_format()
            <faker.providers.python.Provider.pystr_format>`
            for data generation. Argument Groups are used to pass arguments
            to the provider methods. Both ``header`` and ``data_columns`` must
            be of the same length.
        :param num_rows: The ``num_rows`` argument controls how many rows of
            data to generate, and the ``include_row_ids`` argument may be set
            to ``True`` to include a sequential row ID column.
        :param content: File content. Might contain dynamic elements, which
            are then replaced by correspondent fixtures.
        :param encoding: Encoding.
        :param format_func: Callable responsible for formatting template
            strings.
        :param raw: If set to True, return `BytesValue` (binary content of
            the file). Otherwise, return `StringValue` (path to the saved
            file).
        :return: Relative path (from root directory) of the generated file
            or raw content of the file.
        """

        if storage is None:
            storage = FileSystemStorage()

        filename = storage.generate_filename(
            extension=self.extension,
            prefix=prefix,
            basename=basename,
        )

        if self.generator is None:
            self.generator = Faker()

        if content is None:
            data_columns = (
                data_columns if data_columns else DEFAULT_XML_DATA_COLUMNS
            )
            root = ET.Element(root_element)
            for _ in range(num_rows):
                row = ET.SubElement(root, row_element)
                for col_name, col_template in data_columns.items():
                    row.append(
                        self._generate_xml_element(
                            col_name, col_template, format_func=format_func
                        )
                    )
            content = ET.tostring(root, encoding="utf-8").decode("utf-8")
        else:
            content = format_func(self.generator, content)

        data = {"content": content, "filename": filename, "storage": storage}

        if raw:
            raw_content = BytesValue(content.encode("utf8"))
            raw_content.data = data
            return raw_content

        storage.write_text(filename, content, encoding=encoding)

        file_name = StringValue(storage.relpath(filename))
        file_name.data = data
        FILE_REGISTRY.add(file_name)
        return file_name



from django.db import models

__all__ = ("Upload",)


class Upload(models.Model):
    """Upload model."""

    name = models.CharField(max_length=255, unique=True)
    description = models.TextField(null=True, blank=True)
    file = models.FileField()

    class Meta:
        verbose_name = "Upload"
        verbose_name_plural = "Upload"

    def __str__(self: "Upload") -> str:
        return self.name  # noqa
